import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.simulate.peaks import peaks

runner = CliRunner()
app = typer.Typer()
app.command()(peaks)

EXPECTED = """ # noqa: E501
save_nef_nmr_spectrum_synthetic_N_HSQC
   _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
   _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_synthetic_N_HSQC
   _nef_nmr_spectrum.num_dimensions             2
   _nef_nmr_spectrum.chemical_shift_list        .
   _nef_nmr_spectrum.experiment_classification  H[N]
   _nef_nmr_spectrum.experiment_type            '15N HSQC/HMQC'

   loop_
      _nef_spectrum_dimension.dimension_id
      _nef_spectrum_dimension.axis_unit
      _nef_spectrum_dimension.axis_code
      _nef_spectrum_dimension.spectrometer_frequency
      _nef_spectrum_dimension.spectral_width
      _nef_spectrum_dimension.value_first_point
      _nef_spectrum_dimension.folding
      _nef_spectrum_dimension.absolute_peak_positions
      _nef_spectrum_dimension.is_acquisition

     1   ppm   15N   60.833    24.487   125.828   none   true   false
     2   ppm   1H    600.123   0.606    8.535     none   true   true

   stop_

   loop_
      _nef_spectrum_dimension_transfer.dimension_1
      _nef_spectrum_dimension_transfer.dimension_2
      _nef_spectrum_dimension_transfer.transfer_type
      _nef_spectrum_dimension_transfer.is_indirect

     1   2   onebond   false

   stop_

   loop_
      _nef_peak.index
      _nef_peak.peak_id
      _nef_peak.chain_code_1
      _nef_peak.sequence_code_1
      _nef_peak.residue_name_1
      _nef_peak.atom_name_1
      _nef_peak.chain_code_2
      _nef_peak.sequence_code_2
      _nef_peak.residue_name_2
      _nef_peak.atom_name_2
      _nef_peak.position_1
      _nef_peak.position_uncertainty_1
      _nef_peak.position_2
      _nef_peak.position_uncertainty_2
      _nef_peak.height
      _nef_peak.height_uncertainty
      _nef_peak.volume
      _nef_peak.volume_uncertainty

     1   1   '#18'   @19   .     N   '#18'   @19   .     H   123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     2   2   @-      @9    .     N   @-      @9    .     H   124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     3   3   A       10    Ala   N   A       10    Ala   H   102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .

   stop_

save_
""".replace(
    NOQA_E501, ""
)


INPUT_UBI_SHORT_NEF = read_test_data("ubiquitin_short.nef", __file__)


# noinspection PyUnusedLocal
def test_ubi_short():

    result = run_and_report(app, [], input=INPUT_UBI_SHORT_NEF)

    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_N_HSQC")
    assert_lines_match(EXPECTED, new_peaks_frame)


def test_ubi_short_different_frame_name():
    result = run_and_report(
        app, ["--name-template", "wibble_{spectrum}"], input=INPUT_UBI_SHORT_NEF
    )

    isolate_frame(result.stdout, "nef_nmr_spectrum_wibble_N_HSQC")


def test_ubi_short_different_spectrometer_frequency():
    result = run_and_report(
        app, ["--spectrometer-frequency", "800"], input=INPUT_UBI_SHORT_NEF
    )

    EXPECTED_800 = EXPECTED.replace("60.833", "81.095")
    EXPECTED_800 = EXPECTED_800.replace("600.123", "800.000")
    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_N_HSQC")
    assert_lines_match(EXPECTED_800, new_peaks_frame)


EXPECTED_HNCA = """ # noqa: E501
save_nef_nmr_spectrum_synthetic_HNCA
   _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
   _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_synthetic_HNCA
   _nef_nmr_spectrum.num_dimensions             3
   _nef_nmr_spectrum.chemical_shift_list        .
   _nef_nmr_spectrum.experiment_classification  H[N[CA]]
   _nef_nmr_spectrum.experiment_type            HNCA

   loop_
      _nef_spectrum_dimension.dimension_id
      _nef_spectrum_dimension.axis_unit
      _nef_spectrum_dimension.axis_code
      _nef_spectrum_dimension.spectrometer_frequency
      _nef_spectrum_dimension.spectral_width
      _nef_spectrum_dimension.value_first_point
      _nef_spectrum_dimension.folding
      _nef_spectrum_dimension.absolute_peak_positions
      _nef_spectrum_dimension.is_acquisition

         1   ppm   13C   150.933   26.023   63.331    none   true   false
         2   ppm   15N   60.833    29.987   127.078   none   true   false
         3   ppm   1H    600.123   4.400    9.311     none   true   true

   stop_

   loop_
      _nef_spectrum_dimension_transfer.dimension_1
      _nef_spectrum_dimension_transfer.dimension_2
      _nef_spectrum_dimension_transfer.transfer_type
      _nef_spectrum_dimension_transfer.is_indirect

     1   2   onebond   false
     2   3   onebond   false

   stop_

   loop_
      _nef_peak.index
      _nef_peak.peak_id
      _nef_peak.chain_code_1
      _nef_peak.sequence_code_1
      _nef_peak.residue_name_1
      _nef_peak.atom_name_1
      _nef_peak.chain_code_2
      _nef_peak.sequence_code_2
      _nef_peak.residue_name_2
      _nef_peak.atom_name_2
      _nef_peak.chain_code_3
      _nef_peak.sequence_code_3
      _nef_peak.residue_name_3
      _nef_peak.atom_name_3
      _nef_peak.position_1
      _nef_peak.position_uncertainty_1
      _nef_peak.position_2
      _nef_peak.position_uncertainty_2
      _nef_peak.position_3
      _nef_peak.position_uncertainty_3
      _nef_peak.height
      _nef_peak.height_uncertainty
      _nef_peak.volume
      _nef_peak.volume_uncertainty

     1    1    '#18'   @19     .     CA   '#18'   @19    .     N   '#18'   @19    .     H   55.19626104   .              123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     2    2    '#18'   @19-1   .     CA   '#18'   @19    .     N   '#18'   @19    .     H   56.6890552    0.0499432144   123.3654109   0.1052522259   8.507345563   0.003149388014   1000000   .   1000000   .
     3    3    @-      @-9     .     CA   @-      @-9    .     N   @-      @-9    .     H   58.45044906   .              125.7144569   0.1199761464   8.956588347   0.0027175082     1000000   .   1000000   .
     4    4    @-      @-9-1   .     CA   @-      @-9    .     N   @-      @-9    .     H   62.14832291   0.0266869088   125.7144569   0.1199761464   8.956588347   0.0027175082     1000000   .   1000000   .
     5    5    @-      @9      .     CA   @-      @9     .     N   @-      @9     .     H   57.45044906   .              124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     6    6    @-      @9-1    .     CA   @-      @9     .     N   @-      @9     .     H   61.14832291   0.0266869088   124.7144569   0.1199761464   7.956588347   0.0027175082     1000000   .   1000000   .
     7    7    A       -110    Ala   CA   A       -110   Ala   N   A       -110   Ala   H   38.49098528   .              98.4535755    0.0925289285   5.11125646    0.0019229051     1000000   .   1000000   .
     8    8    A       -111    Ala   CA   A       -110   Ala   N   A       -110   Ala   H   48.49098528   .              98.4535755    0.0925289285   5.11125646    0.0019229051     1000000   .   1000000   .
     9    9    A       -11     Ala   CA   A       -10    Ala   N   A       -10    Ala   H   53.49098528   .              103.4535755   0.0925289285   9.11125646    0.0019229051     1000000   .   1000000   .
     10   10   A       9       Ala   CA   A       10     Ala   N   A       10     Ala   H   52.49098528   .              102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .
     11   11   A       110     Ala   CA   A       110    Ala   N   A       110    Ala   H   42.49098528   .              102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .
     12   12   A       109     Ala   CA   A       110    Ala   N   A       110    Ala   H   52.49098528   .              102.4535755   0.0925289285   8.11125646    0.0019229051     1000000   .   1000000   .

   stop_

save_

""".replace(
    NOQA_E501, ""
)


def test_ubi_short_triple():
    test_data = read_test_data("ubiquitin_short_triple.nef", __file__)

    result = run_and_report(app, ["HNCA"], input=test_data)

    new_peaks_frame = isolate_frame(result.stdout, "nef_nmr_spectrum_synthetic_HNCA")
    assert_lines_match(EXPECTED_HNCA, new_peaks_frame)



import typer
from fyeah import f

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.talos.importers.order_parameters import order_parameters

app = typer.Typer()
app.command()(order_parameters)

EXPECTED = """
save_nefpls_order_data_A_N15_H1
   _nefpls_order_data.sf_category         nefpls_order_data
   _nefpls_order_data.sf_framecode        nefpls_order_data_A_N15_H1
   _nefpls_order_data.data_type           model_free
   _nefpls_order_data.relaxation_atom_id  2
   _nefpls_order_data.source              estimate
   _nefpls_order_data.diffusion_model     .

   loop_
      _nefpls_order_values.index
      _nefpls_order_values.chain_code_1
      _nefpls_order_values.sequence_code_1
      _nefpls_order_values.residue_name_1
      _nefpls_order_values.atom_name_1
      _nefpls_order_values.chain_code_2
      _nefpls_order_values.sequence_code_2
      _nefpls_order_values.residue_name_2
      _nefpls_order_values.atom_name_2
      _nefpls_order_values.s2
      _nefpls_order_values.s2_err

     1   A   1   MET   N   A   1   MET   H   0.884   .
     2   A   2   GLN   N   A   2   GLN   H   0.892   .
     3   A   3   ILE   N   A   3   ILE   H   0.903   .
     4   A   4   PHE   N   A   4   PHE   H   0.911   .

   stop_

save_


""".replace(
    NOQA_E501, ""
)


def test_ss_4():

    pred_4_path = path_in_test_data(__file__, "predS2_4.tab")
    pred_4_nef_path = path_in_test_data(__file__, "pred_4_seq.nef")

    STREAM = read_test_data(pred_4_nef_path)

    result = run_and_report(
        app,
        [
            pred_4_path,
        ],
        input=STREAM,
    )

    phi_psi = isolate_frame(result.stdout, "nefpls_order_data_A_N15_H1")

    PATCHED_EXPECTED = f(EXPECTED)
    assert_lines_match(PATCHED_EXPECTED, phi_psi)



import pytest

from nef_pipelines.tools.fit.fit_lib import (
    NEFPLSFitLibInconsistentSeriesDataError,
    RelaxationSeriesValues,
    _combine_relaxation_series,
)

SPECTRA_1 = ["spectrum_A", "spectrum_B"]
PEAK_IDS_1 = [10, 20]
VARIABLE_VALUES_1 = [
    0.1,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6,
    0.7,
    0.8,
    0.9,
    1.0,
]
VALUES_1 = [100.0, 200.0, 300.0, 400.0, 500.0, 600.0, 700.0, 800.0, 900.0, 1000.0]

SPECTRA_2 = ["spectrum_C", "spectrum_D"]
PEAK_IDS_2 = [30, 40]
VALUES_2 = [
    1100.0,
    1200.0,
    1300.0,
    1400.0,
    1500.0,
    1600.0,
    1700.0,
    1800.0,
    1900.0,
    2000.0,
]
VARIABLE_VALUES_2 = [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 1.0]

EXPECTED_SPECTRA = [*SPECTRA_1, *SPECTRA_2]
EXPECTED_PEAK_IDS = [*PEAK_IDS_1, *PEAK_IDS_2]
EXPECTED_VALUES = [*VALUES_1, *VALUES_2]
EXPECTED_VARIABLE_VALUES = [*VARIABLE_VALUES_1, *VARIABLE_VALUES_2]


class TestCombineRelaxationSeries:
    """Test cases for _combine_relaxation_series function."""

    def test_successful_combination_basic(self):
        """Test successful combination of two basic RelaxationSeriesValues."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,
            peak_ids=PEAK_IDS_1,
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1,
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2,
            peak_ids=PEAK_IDS_2,
            variable_values=VARIABLE_VALUES_2,
            values=VALUES_2,
        )

        result = _combine_relaxation_series(series1, series2)

        # Verify result is the correct type
        assert isinstance(result, RelaxationSeriesValues)

        # Verify all fields are combined as blocks (series1 first, then series2)
        assert result.spectra == EXPECTED_SPECTRA
        assert result.peak_ids == EXPECTED_PEAK_IDS
        assert result.variable_values == EXPECTED_VARIABLE_VALUES
        assert result.values == EXPECTED_VALUES

    def test_inconsistent_spectra_length_error(self):
        """Test error when spectra field lengths don't match between series."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,  # length 2
            peak_ids=PEAK_IDS_1,
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1,
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2[:1],  # length 1 - mismatch!
            peak_ids=PEAK_IDS_2,
            variable_values=VARIABLE_VALUES_2,
            values=VARIABLE_VALUES_2,
        )

        with pytest.raises(
            NEFPLSFitLibInconsistentSeriesDataError,
            match=r"Inconsistent field lengths for spectra: series1 has 2, series2 has 1",
        ):
            _combine_relaxation_series(series1, series2)

    def test_inconsistent_peak_ids_length_error(self):
        """Test error when peak_ids field lengths don't match between series."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,  # length 2
            peak_ids=PEAK_IDS_1,
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1,
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2,  # length 1 - mismatch!
            peak_ids=PEAK_IDS_2[:1],
            variable_values=VARIABLE_VALUES_2,
            values=VARIABLE_VALUES_2,
        )

        with pytest.raises(
            NEFPLSFitLibInconsistentSeriesDataError,
            match=r"Inconsistent field lengths for peak_ids: series1 has 2, series2 has 1",
        ):
            _combine_relaxation_series(series1, series2)

    def test_inconsistent_variable_values_length_error(self):
        """Test error when variable_values field lengths don't match between series."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,  # length 2
            peak_ids=PEAK_IDS_1,
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1,
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2,  # length 1 - mismatch!
            peak_ids=PEAK_IDS_2,
            variable_values=VARIABLE_VALUES_2[:5],
            values=VARIABLE_VALUES_2,
        )

        with pytest.raises(
            NEFPLSFitLibInconsistentSeriesDataError,
            match=r"Inconsistent field lengths for variable_values: series1 has 10, series2 has 5",
        ):
            _combine_relaxation_series(series1, series2)

    def test_inconsistent_values_length_error(self):
        """Test error when values field lengths don't match between series."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,  # length 2
            peak_ids=PEAK_IDS_1,
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1,
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2,  # length 1 - mismatch!
            peak_ids=PEAK_IDS_2,
            variable_values=VARIABLE_VALUES_2,
            values=VARIABLE_VALUES_2[:5],
        )

        with pytest.raises(
            NEFPLSFitLibInconsistentSeriesDataError,
            match=r"Inconsistent field lengths for values: series1 has 10, series2 has 5",
        ):
            _combine_relaxation_series(series1, series2)

    def test_multiple_field_length_mismatches(self):
        """Test that the first field mismatch is reported when multiple fields have mismatches."""
        series1 = RelaxationSeriesValues(
            spectra=SPECTRA_1,  # length 2
            peak_ids=PEAK_IDS_1[:5],
            variable_values=VARIABLE_VALUES_1,
            values=VALUES_1[:5],
        )

        series2 = RelaxationSeriesValues(
            spectra=SPECTRA_2[:1],  # length 1 - mismatch!
            peak_ids=PEAK_IDS_2,
            variable_values=VARIABLE_VALUES_2[:5],
            values=VARIABLE_VALUES_2,
        )

        # Should report the first field that has a mismatch (spectra)
        with pytest.raises(
            NEFPLSFitLibInconsistentSeriesDataError,
            match=r"Inconsistent field lengths for spectra: series1 has 2, series2 has 1",
        ):
            _combine_relaxation_series(series1, series2)



import string
from enum import auto
from typing import List, Sequence, Union

from strenum import UppercaseStrEnum

ROUND_1H_FREQUENCIES = [i * 100.0 for i in range(4, 20)]
TYPICAL_1H_FREQUENCIES = [*ROUND_1H_FREQUENCIES, 750.0, 850.0, 950.0]

# from https://www.kherb.io/docs/nmr_table.html
# The values are taken from the IAEA's table of recommended nuclear magnetic moments and have been converted to
# gyromagnetic ratios using the 2018 CODATA recommended values for the nuclear magneton and Planck's constant.
# However, the value for hydrogen has been replaced with the CODATA value. The number in parentheses represents
# the one-sigma (1σ) uncertainty in the last two digits of the value provided.
# this is agreement with the values in the Cavanagh et als book (2nd edition)
# 42.577478518 * 2 * 3.1415927 / 100 = 2.675221874410521 vs 2.6752 in Cavanagh within reasonable errors
# the value from NIST is 2.675 221 8708 (12) s-1 T-1 which is in agreement with the value in Cavanagh
MAGNETOGYRIC_RATIO_1H = 42.577478518  # uncertainty (18)  MHz/T


class Isotope(UppercaseStrEnum):
    H1 = auto()
    H2 = auto()
    H3 = auto()
    N15 = auto()
    C13 = auto()
    O17 = auto()
    F19 = auto()
    P31 = auto()

    def __str__(self):
        element = self.name.rstrip(string.digits)
        isotope_number = self.name[len(element) :]
        return f"{isotope_number}{element}"


# fmt: off
GAMMA_RATIOS = {
    # where did i get these from (not wikipedia or the bruker handbook)? is that an authoritative source?
    # Nuc	        ratio	        # (MHz⋅T−1)
    Isotope.H1:     1.0,            # 42.58
    Isotope.H2:     0.153508386,    # 6.54
    Isotope.H3:     1.066643718,    # 45.42
    Isotope.C13:    0.251503855,    # 10.71
    Isotope.N15:    0.101368145,    # -4.32 - should really be negative !
    Isotope.O17:    0.135564627,    # -5.77 - should really be negative !
    Isotope.F19:    0.94129576,     # 40.08
    Isotope.P31:    0.404791467,    # 17.24
}

ATOM_TO_ISOTOPE = {
    # Nuc   Isotope
    "H": Isotope.H1,
    "D": Isotope.H2,
    "T": Isotope.H3,
    "C": Isotope.C13,
    "N": Isotope.N15,
    "O": Isotope.O17,
    "F": Isotope.F19,
    "P": Isotope.P31,
}

CODE_TO_ISOTOPE = {
    # Nuc   Isotope
    "1H": Isotope.H1,
    "2H": Isotope.H2,
    "3H": Isotope.H3,
    "13C": Isotope.C13,
    "15N": Isotope.N15,
    "17O": Isotope.O17,
    "19F": Isotope.F19,
    "31P": Isotope.P31,
}
# fmt: on


def convert_isotopes(
    isotopes: Union[Union[str, Isotope], Sequence[Union[str, Isotope]]]
) -> List[Union[Isotope, str]]:
    """
    given a list or single values of either isotope names as strings or Isotopes convert them to a
    list of Isotopes where possible. Unrecognised isotopes are passed throughnas strings. Isotope strings
    can be either C13 or 13C for known  isotopes...

    isotopes: either one or a list of isotope name or Isotopes

    result: a list of Isotopes or strs for Isotopes that weren't recognised
    """
    if isinstance(isotopes, (str, Isotope)):
        isotopes = [
            isotopes,
        ]

    result = []
    isotope_values = [member.value for member in Isotope]
    for isotope in isotopes:
        if isotope in CODE_TO_ISOTOPE:
            result.append(CODE_TO_ISOTOPE[isotope])
        elif isotope in isotope_values:
            result.append(Isotope(isotope))
        else:
            result.append(isotope)

    return result



import sys
from argparse import Namespace
from collections.abc import MutableMapping
from enum import auto
from fnmatch import fnmatch
from itertools import zip_longest
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

# from pandas import DataFrame
from pynmrstar import Entry, Loop, Saveframe
from pynmrstar.definitions import STR_CONVERSION_DICT
from pynmrstar.exceptions import ParsingError
from strenum import LowercaseStrEnum

from nef_pipelines.lib import util
from nef_pipelines.lib.constants import NEF_PIPELINES
from nef_pipelines.lib.globals_lib import set_global
from nef_pipelines.lib.util import (
    exit_error,
    fixup_metadata,
    get_version,
    is_float,
    is_int,
    running_in_pycharm,
    script_name,
    warn,
)

NEF_RELAXATION_VERSION = "0.1.0a"

DEFAULT_INCREMENT_SEPARATOR = "::"
DEFAULT_INCREMENT_END = ""

NEF_TRUE = "true"
NEF_FALSE = "false"
NEF_NONE = "none"

NEF_CATEGORY_ATTR = "__NEF_CATEGORY__"
NEF_MOLECULAR_SYSTEM = "nef_molecular_system"
NEF_METADATA = "nef_nmr_meta_data"

UNUSED = "."
UNDERSCORE = "_"

# this stops pynmrstar failing on empty strings
STR_CONVERSION_DICT[""] = UNUSED


class BadNefFileException(Exception):
    """
    marker class for star files that don't conform to NEF conventions, e.g. files with multiple singletons
    as an example the molecular system should be a singleton
    ths doesn't imply the file has invalid star syntax, the star syntax will be valid
    """

    pass


class NoSuchColumnException(Exception):
    pass


# what the selection type is for Star SaveFrames
class SelectionType(LowercaseStrEnum):
    NAME = auto()
    CATEGORY = auto()
    ANY = auto()

    # see https://github.com/irgeek/StrEnum/issues/9
    def _cmp_values(self, other):
        return self.value, str(other).upper()


SELECTORS_LOWER = ", ".join(
    [selector.lower() for selector in SelectionType.__members__]
)


class PotentialTypes(LowercaseStrEnum):
    UNDEFINED = auto()
    LOG_HARMONIC = "log-harmonic"
    PARABOLIC = auto()
    SQUARE_WELL_PARABOLIC = "square-well-parabolic"
    SQUARE_WELL_PARABOLIC_LINEAR = "square-well-parabolic-linear"
    UPPER_BOUND_PARABOLIC = "upper-bound-parabolic"
    LOWER_BOUND_PARABOLIC = "lower-bound-parabolic"
    UPPER_BOUND_PARABOLIC_LINEAR = "upper-bound-parabolic-linear"
    LOWER_BOUND_PARABOLIC_LINEAR = "lower-bound-parabolic-linear "

    # see https://github.com/irgeek/StrEnum/issues/9
    def _cmp_values(self, other):
        return self.value, str(other).upper()


POTENTIAL_TYPES_LOWER = ", ".join(
    [selector.lower() for selector in PotentialTypes.__members__]
)

# currently disabled as they add a dependency on pandas and numpy
# def loop_to_dataframe(loop: Loop) -> DataFrame:
#     """
#     convert a pynmrstar Loop to a pandas DataFrame. Note the Loop category is
#     saved in the dataframe's attrs['__NEF_CATEGORY__']
#
#     :param loop: the pynmrstar Loop
#     :return: a pandas DataFrame
#     """
#     tags = loop.tags
#     data = DataFrame()
#
#     # note this strips the preceding _
#     data.attrs[NEF_CATEGORY_ATTR] = loop.category[1:]
#
#     for tag in tags:
#         if tag != "index":
#             data[tag] = loop.get_tag(tag)
#
#     return data
#
#
# def dataframe_to_loop(frame: DataFrame, category: str = None) -> Loop:
#     """
#     convert a pandas DataFrame to a pynmrstar Loop
#
#     :param frame: the pandas DataFrame
#     :param category: the star category note this will override any category stored in attrs
#     :return: the new pynmrstar Loop
#     """
#     loop = Loop.from_scratch(category=category)
#     loop_data = {}
#     for column in frame.columns:
#         loop.add_tag(column)
#         loop_data[column] = list(frame[column])
#
#     loop.add_data(loop_data)
#
#     if NEF_CATEGORY_ATTR in frame.attrs and not category:
#         loop.set_category(frame.attrs[NEF_CATEGORY_ATTR])
#     elif category:
#         loop.set_category(category)
#
#     return loop


def get_frame_id(frame: Saveframe) -> str:
    """
    given a frame get its id (the characters after the category and a separating _ charctacter)
    :param frame: the frame to get the name of
    :return: the id of the frame
    """
    return frame.name[len(frame.category) + 1 :]


def get_frame_ids(frames_or_entry: Union[List[Saveframe], Entry]) -> List[str]:
    """
    given a list of frames get theirs id (the characters after the category and a separating _ charctacter)
    :param frames_or_entry: a list of frames or an entty
    :return: a list of ids in the same order
    """

    return list([get_frame_id(frame) for frame in frames_or_entry])


def select_frames_by_name(
    frames: Union[List[Saveframe], Entry],
    name_selectors: Union[List[str], str],
    exact=False,
) -> Tuple[Saveframe, ...]:
    """
    select frames  by names and wild cards, to avoid typing *s on the command line the match is greedy by default,
    if an exact match is not found for one of the frames first time we search again with all the name selectors
    turned into wild cards by surrounding them with the * as a fence so name_selector-> *name_selector*

    :param frames: the list of frames or entry to search
    :param name_selectors: a single string or list of strings to use to match frame names, selectors can contain
                           wild cards used by pythons fnmatch
    :param exact: do exact matching and don't search again with wildcards added if no frames are selected
    :return: a list or matching frames
    """

    def match_frames(
        frames: List[Saveframe], name_selectors: List[str]
    ) -> Dict[Any, Saveframe]:
        result = {}

        for frame in frames:
            for selector in name_selectors:

                if fnmatch(frame.name, selector):
                    # frames aren't hashable and so can't be saved in a set
                    # but names should be unique
                    result[frame.name] = frame

        return result

    if isinstance(name_selectors, str):
        name_selectors = [
            name_selectors,
        ]

    result = match_frames(frames, name_selectors)

    if not exact and len(result) == 0:

        name_selectors = [f"*{selector}*" for selector in name_selectors]
        result = match_frames(frames, name_selectors)

    return tuple(result.values())


# refactor to two functions one of which gets a TextIO
def create_entry_from_stdin() -> Optional[Entry]:
    """
    read a star file entry from stdin or return None

    can throw a ParsingError from PyNMRStar

    :return: a star file entry or None
    """

    try:
        entry = None
        if not sys.stdin.isatty() or running_in_pycharm():
            stdin_lines = sys.stdin.read()
            if stdin_lines is None:
                lines = ""
            else:
                lines = "".join(stdin_lines)

            if len(lines.strip()) != 0:
                entry = Entry.from_string(lines)
    except ParsingError as e:
        raise BadNefFileException(str(e)) from e

    if entry:
        entry.source = "-"

    return entry


def read_file_or_exit(file_path: Path) -> List[str]:
    """
    read the contents of a file or exit with error

    :param file_path: the path to the file
    :return: a list of strings from the file
    """
    try:
        with open(file_path) as fh:
            result = fh.readlines()
    except IOError as e:
        msg = f"""\
            failed to read from {file_path} because {e}
        """
        exit_error(msg)

    return result


# refactor to two functions one of which gets a TextIO
def read_entry_from_stdin_or_exit() -> Entry:
    """
    read a star file entry from stdin or exit with an error message
    :return: a star file entry
    """

    if sys.stdin.isatty():
        exit_error("you appear to be reading from an empty stdin")

    try:
        stdin_lines = sys.stdin.read()
        if stdin_lines is None:
            lines = [
                "",
            ]
        lines = "".join(stdin_lines)
    except IOError as e:
        exit_error(
            f"failed to read stdin because: {e}",
            e,
        )

    if len(lines.strip()) == 0:
        exit_error("stdin is empty")

    try:
        entry = Entry.from_string(lines)

    except ParsingError as e:
        exit_error(
            f"failed to read nef entry from stdin because the NEF parser replied: {e}",
            e,
        )

    entry.source = "-"

    return entry


class RowDict(MutableMapping):
    def __init__(self, loop: Loop, row: int, convert: bool):
        super().__init__()
        self._data = row
        self._convert = convert
        self._tag_to_index = {tag: i for i, tag in enumerate(loop.tags)}
        self._tag_types = {
            tag: type(value) for tag, value in zip(loop.tags, self._data)
        }

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._tag_to_index.keys())

    def __delitem__(self, key):
        self._data[self._tag_to_index[key]] = UNUSED

    def __getitem__(self, key):
        index = self._tag_to_index[key]
        value = self._data[index]
        if self._convert:
            value = do_reasonable_type_conversions(value)
        return value

    def __setitem__(self, key, value):
        index = self._tag_to_index[key]
        if self._tag_types[key] == str and not isinstance(value, str):
            value = str(value)
        self._data[index] = value

    def __repr__(self):
        row = {key: self._data[index] for key, index in self._tag_to_index.items()}
        return str(row)

    def __str__(self):
        return self.__repr__()

    # def __getattribute__(self, item):
    #     print(item)
    #     try:
    #         return self[item]
    #     except Exception:
    #         return object.__getattribute__(self, item)
    #
    # def __setattr__(self, item, value):
    #     if hasattr(self, item):
    #         self[item] = value
    #     else:
    #         return super().__setattr__(item, value)
    #
    # def __delattr__(self, key):
    #     self.__delitem__(key)


# TODO we should examine columns for types not individual rows entries
# TODO we should rename this loop_row_iter
def loop_row_dict_iter(
    loop: Loop, convert: bool = True
) -> Iterator[Dict[str, Union[str, int, float]]]:
    """
    create an iterator that loops over the rows in a star file Loop as dictionaries, by default sensible
    conversions from strings to ints and floats are made
    :param loop: the Loop
    :param convert: try to convert values to ints or floats if possible [default is True]
    :return: iterator of rows as dictionaries
    """

    if not isinstance(loop, Loop):
        msg = f"""\
            loop must be of type Loop you provided a {loop.__class__.__name__}"
            value: {loop}
        """
        raise Exception(msg)

    for i, row in enumerate(loop):
        result = RowDict(loop, row, convert)
        yield result


def do_reasonable_type_conversions(value: str) -> Union[str, float, int]:
    """
    do reasonable type conversion from str to int or float
    :param value: the string to convert
    :return: value converted from str to int or float if possible
    """

    # only apply conversions to strings otherwise you get nice unexpected conversions like float -> int
    if isinstance(value, str):
        if is_int(value):
            value = int(value)
        elif is_float(value):
            value = float(value)
        elif value.lower() == NEF_FALSE:
            value = False
        elif value.lower() == NEF_TRUE:
            value = True
    return value


def loop_row_namespace_iter(loop: Loop, convert: bool = True) -> Iterator[Namespace]:
    """
    create an iterator that loops over the rows in a star file Loop as Namespaces, by default sensible
    conversions from strings to ints and floats are made
    :param loop: thr Loop
    :param convert: try to convert values to ints or floats if possible [default is True]
    :return: iterator of rows as dictionaries

    NOTE: This function is effectively replaced by loop_row_dict_iter, which returns a RowDict object
          with support for attribute lookup
    """
    for row in loop_row_dict_iter(loop, convert=convert):
        # yield row needs more tests modified
        yield Namespace(**row)


# TODO this partially overlaps with select_frames_by_name in this file, combine and simplify!
def select_frames(
    entry: Entry,
    predicate: Union[str, List[str]],
    selector_type: SelectionType = SelectionType.ANY,
) -> List[Saveframe]:
    """
    select a list of frames by name of either category or name, if no predicates are provided all frames are selected

    :param entry: the entry in which frames are looked for
    :param predicate: a string or list of strings to use as filters as defined by fnmatch
    :param selector_type: the matching type frame.name or frame.category or both [default, search order
           frame.name frame.category]
    :return: a list of selected saveframes
    """

    if isinstance(predicate, str):
        predicate = [
            predicate,
        ]
    if not predicate:
        predicate = ["*"]

    star_filters = [f"*{filter}*" for filter in predicate]
    predicate = list(predicate)
    predicate.extend(star_filters)

    result = {}
    for frame in entry.frame_dict.values():

        if frame.category is not None:
            accept_frame_category = any(
                [fnmatch(frame.category, filter) for filter in predicate]
            )
        else:
            accept_frame_category = False
        accept_frame_name = any([fnmatch(frame.name, filter) for filter in predicate])

        if (
            selector_type in (SelectionType.NAME, SelectionType.ANY)
            and accept_frame_name
        ):
            result[frame.category, frame.name] = frame

        if (
            selector_type in (SelectionType.CATEGORY, SelectionType.ANY)
            and accept_frame_category
        ):
            result[frame.category, frame.name] = frame

    return list(result.values()) if result else []


def read_entry_from_file_or_stdin_or_exit_error(file: Path) -> Entry:
    """
    read a star entry from stdin or a file or exit.
    note this exits with an error if stdin can't be read because its a terminal

    :param file:
    :return:
    """
    if file is None or file == Path("-"):
        entry = read_entry_from_stdin_or_exit()
    else:
        entry = read_entry_from_file_or_exit_error(file)
    return entry


def read_entry_from_file_or_exit_error(file):
    """
    read a star entry from a file or exit with an error message

    :param file: a file on the file system
    :return: a new Entry read from the file
    """

    file = Path(file)

    if not file.exists() or not file.is_file():
        exit_error(f"the file {file} doesn't exist or isn't a file")
    try:
        with open(file) as fh:
            entry = Entry.from_file(fh)

    except IOError as e:
        exit_error(f"couldn't read from the file {file}", e)

    return entry


# TODO: should be used as underpinnings for other related functions here
def read_or_create_entry_exit_error_on_bad_file(
    file: Path, entry_name: str = "nef"
) -> Entry:
    """
    read a star entry from stdin or create a new one. If a badly formatted Star file is read exit error...

    :param file: a file whih can be either on the file system or stding Path('-')
    :return: an Entry a new Entry is created if no entry can be read
    """

    entry = None
    try:
        if file is None or file == Path("-"):
            entry = create_entry_from_stdin()
            if entry and entry.entry_id == "nefpls_globals":
                entry.entry_id = entry_name
        else:
            try:
                with open(file) as fh:
                    entry = Entry.from_file(fh)
                    _parse_globals(entry)

            except IOError as e:
                msg = f"""
                    while reading the file {util.get_display_file_name(file)} the file wasn't readable
                    error from operating system was {str(e)}
                """
                exit_error(msg)
    except BadNefFileException as e:
        msg = f"""
            while reading the file {util.get_display_file_name(file)} the format was bad
            error from star file parser was {str(e)}
        """
        exit_error(msg)

    if entry is None:
        entry = Entry.from_scratch(entry_name)

    return entry


def _parse_globals(entry: Entry):

    globals_frames = entry.get_saveframes_by_category("nefpls_globals")
    for globals_frame in globals_frames:

        values = {name: globals_frame.get_tag(name) for name in globals_frame.tags}
        values = {
            name: do_reasonable_type_conversions(value)
            for name, value in values.items()
        }
        for key, value in values.items():
            set_global(key, value)


def file_name_path_to_frame_name(path: str) -> str:
    """
    convert a filename or complete path to a name suitable for use as a frame name, this does the following mappings
        . -> _
        - -> _

    :param file_name: the filename to convert
    :return: a string suitable for use as a NEF framename
    """
    path = Path(path)

    stem = str(path.stem)
    stem = stem.replace("-", "_")
    stem = stem.replace(".", "_")

    return stem


def molecular_system_from_entry_or_exit(entry: Entry):
    """
    read the nef molecular system [a singleton] from a star entry or exit with a helpful message
    :param entry: The NEF star entry
    :return: the molecular system save frame or None if there isn't one
    """

    molecular_system = None
    try:
        molecular_system = molecular_system_from_entry(entry)
    except BadNefFileException as e:
        exit_error(str(e))

    if not molecular_system:
        exit_error(
            "Couldn't find a molecular system frame it should be labelled 'save_nef_molecular_system'"
        )

    return molecular_system


def molecular_system_from_entry(entry):
    """
    read the nef molecular system [a singleton] from a star entry or raise a BadNefFileException
    :param entry: The NEF star entry
    :return: the molecular system save frame or None if there isn't one
    """
    result = None

    molecular_systems = entry.get_saveframes_by_category(NEF_MOLECULAR_SYSTEM)
    if len(molecular_systems) == 1:
        result = molecular_systems[0]
    elif len(molecular_systems) > 1:
        msg = f"""\
                there must be only one molecular_system i found {len(molecular_systems)}
                frame_names are {', '.join([frame.name for frame in molecular_systems])}
        """
        raise BadNefFileException(msg)

    return result


def add_frames_to_entry(entry: Entry, frames: List[Saveframe]) -> Entry:
    # TODO deal with merging esp wrt to molecular systems and possibly with other information
    # TODO add frame rename and frame delete
    """
    take a set of save frames and  add them to an Entry and update the NEF metadata header

    Args:
        entry: an entry to add the save frames to
        frames: a set of save frames to add, they must have different names to those present already


    Returns:
        the updated entry containing the frames
    """

    fixup_metadata(entry, NEF_PIPELINES, get_version(), script_name(__file__))

    for frame in frames:

        new_frame_name = frame.name

        proposed_new_frame_name = new_frame_name
        offset = ""
        while is_save_frame_name_in_entry(entry, proposed_new_frame_name):
            if offset == "":
                offset = 1
            else:
                offset += 1
            if offset != "":
                proposed_new_frame_name = f"{new_frame_name}_{offset}"

        if proposed_new_frame_name != new_frame_name:

            msg = f"""
                the frame named {new_frame_name} already exists in the stream, the frame name has been updated to
                {proposed_new_frame_name}. if you want to replace the original frame named {new_frame_name} delete
                it first!
            """
            msg = dedent(msg).strip()
            warn(msg)

        frame.name = proposed_new_frame_name

        entry.add_saveframe(frame)

    return entry


def is_save_frame_name_in_entry(entry: Entry, frame_name: str) -> bool:
    """
    check if a save frame name exists in an entry
    :param entry: the entry
    :param frame_name: the frame name
    :return: if a savefame with the name is present
    """
    frame_in_entry = True
    try:
        entry.get_saveframe_by_name(frame_name)
    except KeyError:
        frame_in_entry = False

    return frame_in_entry


def set_column(loop: Loop, column: Union[int, str], values: List[Any]) -> Loop:
    """
    set the values of a column in a loop from a list
    :param loop:  the loop
    :param column: the name or index of the column to modify
    :param values: the values to put into the column
    :return: the loop (same instance)
    """

    column = _ensure_column_is_index(loop, column)

    SENTINEL = "!!SENTINEL!!"
    for row, merit in zip_longest(loop, values, fillvalue=SENTINEL):
        if row is SENTINEL:
            break
        row[column] = merit

    return loop


def _ensure_column_is_index(loop: Loop, column: Union[str, int]) -> int:
    if isinstance(column, str):
        if column in loop.tags:
            column = loop.tag_index(column)
        else:
            msg = f"""
                the column {column} wasn't found in the loop {loop.category}

                the available columns are {', '.join(loop.tags)}
            """
            raise NoSuchColumnException(msg)
    return column


def set_column_to_value(loop: Loop, column: Union[str, int], value: Any) -> Loop:
    """

    set the values of a column in a loop to a single value

    :param loop:  the loop
    :param column: the name or index of the column to modify
    :param values: the values to put
    :return: the loop (same instance)
    """
    column = _ensure_column_is_index(loop, column)
    for row in loop:
        row[column] = value

    return loop


def extract_column(loop: Loop, column: Union[str, int]) -> Loop:
    """
    Extract all the values in the column of a loop into a list

    :param the loop
    :param column: the name or index of the column to modify
    :return: the values in the column as a list in row order
    """
    column = _ensure_column_is_index(loop, column)
    values = []
    for row in loop:
        values.append(row[column])
    return values


def create_nef_save_frame(
    frame_category: str,
    frame_id: str = None,
    increment: str = "",
    separator: str = DEFAULT_INCREMENT_SEPARATOR,
    end: str = DEFAULT_INCREMENT_END,
) -> Saveframe:
    """
    Create a NEF save frame from a category and a frame name, including the tags sf_category and sf_framecode

    :param category: the category including an organisation prefix
    :param frame_id: the id of the frame  (can be None for singletons)
    :param separator: separator between the category and the frame_id (default ::)
    :param end: text after the frame_id (can be used to create a fence  default is the empty string
    :return: a NEF style saveframe
    """

    if frame_id and increment:
        frame_name = f"{frame_category}_{frame_id}{separator}{increment}{end}"
    elif frame_id:
        frame_name = f"{frame_category}_{frame_id}"
    else:
        frame_name = f"{frame_category}"

    frame = Saveframe.from_scratch(frame_name, f"{frame_category}")
    frame.add_tag("sf_category", frame_category)
    frame.add_tag("sf_framecode", frame_name)
    return frame



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrview.importers.peaks import peaks

MOLECULAR_SYSTEM_NMRVIEW = "nef_spectrum_nmrview"

app = typer.Typer()
app.command()(peaks)


EXPECTED_4AA = read_test_data("nmrview_expected_4aa.txt", __file__)


# noinspection PyUnusedLocal
def test_4_peaks():
    peaks_path = path_in_test_data(__file__, "4peaks.xpk")

    sequence_stream = read_test_data("4peaks_seq.nef", __file__)

    result = run_and_report(
        app,
        [
            peaks_path,
        ],
        input=sequence_stream,
    )

    result = isolate_frame(result.stdout, "nef_nmr_spectrum_simnoe")

    assert_lines_match(EXPECTED_4AA, result)


EXPECTED_JOE_CLIC = """\
    save_nef_nmr_spectrum_trosyrdcjoe
        _nef_nmr_spectrum.sf_category          nef_nmr_spectrum
        _nef_nmr_spectrum.sf_framecode         nef_nmr_spectrum_trosyrdcjoe
        _nef_nmr_spectrum.num_dimensions       2
        _nef_nmr_spectrum.chemical_shift_list  .

        loop_
            _nef_spectrum_dimension.dimension_id
            _nef_spectrum_dimension.axis_unit
            _nef_spectrum_dimension.axis_code
            _nef_spectrum_dimension.spectrometer_frequency
            _nef_spectrum_dimension.spectral_width
            _nef_spectrum_dimension.value_first_point
            _nef_spectrum_dimension.folding
            _nef_spectrum_dimension.absolute_peak_positions
            _nef_spectrum_dimension.is_acquisition

            1   ppm   1H   600.052978516    1.0   8.65968   circular   true   .
            2   ppm  15N   60.8100013733    1.0 127.47411   circular   true   .

        stop_

        loop_
            _nef_spectrum_dimension_transfer.dimension_1
            _nef_spectrum_dimension_transfer.dimension_2
            _nef_spectrum_dimension_transfer.transfer_type


        stop_

        loop_
            _nef_peak.index
            _nef_peak.peak_id
            _nef_peak.volume
            _nef_peak.volume_uncertainty
            _nef_peak.height
            _nef_peak.height_uncertainty
            _nef_peak.position_1
            _nef_peak.position_uncertainty_1
            _nef_peak.position_2
            _nef_peak.position_uncertainty_2
            _nef_peak.chain_code_1
            _nef_peak.sequence_code_1
            _nef_peak.residue_name_1
            _nef_peak.atom_name_1
            _nef_peak.chain_code_2
            _nef_peak.sequence_code_2
            _nef_peak.residue_name_2
            _nef_peak.atom_name_2

            1   1   0.254054695368   .   0.0368   .   8.15968   .   126.97411   .   A   2   ALA   H   A   2   ALA   N

        stop_

    save_
"""


# noinspection PyUnusedLocal
def test_joe_bad_sweep_widths():
    peaks_path = path_in_test_data(__file__, "joe_clic.xpk")

    sequence_stream = read_test_data("4peaks_seq.nef", __file__)

    result = run_and_report(
        app,
        [
            peaks_path,
        ],
        input=sequence_stream,
    )

    result = isolate_frame(result.stdout, "nef_nmr_spectrum_trosyrdcjoe")

    assert_lines_match(EXPECTED_JOE_CLIC, result)



from pathlib import Path

import typer

from nef_pipelines.lib.nef_lib import read_or_create_entry_exit_error_on_bad_file
from nef_pipelines.lib.sequence_lib import MoleculeType
from nef_pipelines.lib.util import STDIN
from nef_pipelines.transcoders.fasta.importers.sequence import pipe
from nef_pipelines.transcoders.mars import import_app


@import_app.command(no_args_is_help=True)
def sequence(
    chain_code: str = typer.Option(
        None,
        "--chain",
        help="chain code to use for the imported chains [only one mars only supports one]",
        metavar="<CHAIN-CODE>",
    ),
    start: int = typer.Option(
        1,
        "--start",
        help="first residue number of the sequence",
        metavar="<START>",
    ),
    no_chain_start: bool = typer.Option(None, "--no-chain-start", help=""),
    no_chain_end: bool = typer.Option(None, "--no-chain-end", help=""),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_name: Path = typer.Argument(
        ..., help="input files of type sequence.fasta", metavar="<MARS-SEQUENCE>.fasta"
    ),
):
    """- import mars sequence from fasta file"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    chain_codes = (
        [
            chain_code,
        ]
        if chain_code
        else []
    )
    starts = [
        start,
    ]
    no_chain_starts = (
        [
            no_chain_start,
        ]
        if no_chain_start
        else []
    )
    no_chain_ends = (
        [
            no_chain_end,
        ]
        if no_chain_end
        else []
    )
    molecule_types = [
        MoleculeType.PROTEIN,
    ]
    file_names = [
        file_name,
    ]

    entry = pipe(
        entry,
        chain_codes,
        starts,
        no_chain_starts,
        no_chain_ends,
        molecule_types,
        False,
        file_names,
        file_name.root,
    )

    print(entry)



import string
import sys
from collections import namedtuple
from itertools import zip_longest
from typing import Dict, List, Tuple

from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.sequence_lib import (
    TRANSLATIONS_1_3,
    MoleculeType,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    NewPeak,
    PeakFitMethod,
    Residue,
    SequenceResidue,
    ShiftData,
)
from nef_pipelines.lib.util import exit_error, is_float, strip_characters_left

#
# sparky docs
#
# Make peaks on a spectrum from a peak list file (rp).
#
# Peaks are read from a peak list file and placed on the selected spectrum.
# The file should contain a line for each peak. The line should have an
# assignment followed by chemical shifts for each axis. For example
#
# 	C2H5-G1H1      5.395      6.030
# The assignment can contain ? components. It can omit a group name for a component
# -- the shorthand G1H1'-H2' where the second group is omitted is equivalent to G1H1'-G1H2'.
# The residue name is separated from the atom name by looking for a residue number followed
# by one of the letters H, C, N, Q, or M (upper or lower case). Extra columns become the
# peak note. Peaks for 3-D or 4-D spectra can be read.
#
# what does sparky do with multiple chains?
#
# what does ga mean in the sparky output after a volume
#
# why does sparky sometime have ga after a volume I presume it's a gausiann peak fit?
# comment gst 19/04/2023 is ga for a gaussian peak fit?
#
#  Assignment       w1      w2      Volume     lw1 (hz)   lw2 (hz)
#
#    G16H3'-H8    4.905   8.010   7.15e+06 ga    28.6       20.0
#    G16H4'-H8    4.439   8.013   5.42e+06 ga    35.3       16.9
#  T17H6-G16H8    7.205   8.004   1.68e+06
#  T17H7-G16H8    1.459   8.008   2.09e+07 ga    27.5       24.1
#   T17H2"-H1'    2.509   5.840   4.68e+07 ga    41.2       17.6
#
# what does ta display in a sparky project
# how does sparky cope with complexes where peaks are between chains
# which it doesn't define
#

SparkyAtomLabel = namedtuple("SparkyAtomLabel", "residue_number residue_name atom_name")


def parse_single_assignment(
    assignment: str,
    previous_residue_name: str = None,
    previous_residue_number: str = None,
) -> Tuple[str, str, str]:
    """
    Parse a single sparky assignment of the form ?, G16H4', H8, T17H6, G16H8 into a tuple of residue_number,
    residue_name and atom_name. To allow for processing of abbreviated assignments of the form G16H4'-H8 a previous
    residue number and residue name may be passed in [in this casefor H8  they would be G and 16 to match G16H4']

    :param assignment: the single sparky assignments
    :param previous_residue_name: a previous residue name to use for an abbreviated assignment
    :param previous_residue_number: a previous residue number to use for an abbreviated assignment
    :return: a tuple of the form (residue_number, residue_name , atom_name)
    """

    target = assignment

    # remove a residue name or question mark
    residue_name, target = strip_characters_left(target, string.ascii_letters)
    if len(target) == 0:
        target = residue_name
        residue_name = ""

    if target[0] == "?" and len(target) > 0:
        target = target[1:]
        residue_name = ""

    # remove residue numbers or question mark
    residue_number, target = strip_characters_left(target, string.digits)

    if len(residue_number) == 0 and len(target) != 0 and target[0] == "?":
        target = target[1:]
        residue_number = ""

    if len(target) > 0:
        atom_name = target
    else:
        atom_name = ""

    # residue name is separated from the atom name by looking for a residue number followed
    # by one of the letters H, C, N, Q, or M ergo a residue name that starts with one of these
    # must be wrong, and we have split an atom (though we don't actually need this test?)
    residue_lower = residue_name.lower()
    nmr_atoms_lower = "HCNQM".lower()
    atom_starts_with_nmr_lower = False
    atom_name_lower = atom_name.lower()
    len_atom_name = len(atom_name)
    if len_atom_name > 0 and atom_name_lower[0] in nmr_atoms_lower:
        atom_starts_with_nmr_lower = True

    if (
        not atom_starts_with_nmr_lower
        and residue_name != ""
        and residue_lower in nmr_atoms_lower
    ):
        atom_name = f"{residue_name}{residue_number}{atom_name}"
        residue_name = ""
        residue_number = ""

    if (
        residue_number == ""
        and previous_residue_number is not None
        and residue_name == ""
        and previous_residue_name is not None
        and atom_name != ""
    ):
        residue_number = previous_residue_number
        residue_name = previous_residue_name

    if residue_name == "" and residue_number == "":
        for i, character in enumerate(atom_name):
            if character.lower() in "HCNQM".lower():
                if i > 0 and atom_name[i - 1] in string.digits:
                    residue_name_and_number = atom_name[:i]
                    atom_name = atom_name[i:]

                    residue_name = residue_name_and_number.rstrip(string.digits)

                    residue_number = residue_name_and_number[-len(residue_name) :]

    return SparkyAtomLabel(residue_number, residue_name, atom_name)


def parse_assignments(
    assignments: str,
    chain_code: str,
    sequence: List[SequenceResidue],
    molecule_type: MoleculeType,
    line_info: LineInfo,
    allow_pseudo_atoms=False,
) -> List[AtomLabel]:
    """
    take a sparky assignment of the form ?, G16H4'-H8 or T17H6-G16H8 into atom labels

    :param assignments: a single assignment or a - separated assignment [including abbreviated assignments]
    :param chain_code: the chain code to use
    :param sequence: the sequence to get residue names from and to check the read residue names agains
    :param molecule_type: the type of the molecule DNA RNS protein etc.
    :param line_info: file and line  information for error reporting
    :param allow_pseudo_atoms: if true ignore residue names which aren't in the molecule_type
    :return: a list of AtomLabels
    """

    residue_name_lookup = sequence_to_residue_name_lookup(sequence)

    residue_name_translations = TRANSLATIONS_1_3[molecule_type]

    fields = assignments.split("-")

    result = []
    last_sequence_code = None
    last_residue_name = None

    for field in fields:

        sequence_code, residue_name, atom_name = parse_single_assignment(
            field,
            previous_residue_number=last_sequence_code,
            previous_residue_name=last_residue_name,
        )

        if sequence_code == "":
            sequence_code = UNUSED
        if residue_name == "":
            residue_name = UNUSED

        if residue_name_translations is not None and residue_name is not UNUSED:

            if residue_name in residue_name_translations:
                translated_residue_name = residue_name_translations[residue_name]
            elif allow_pseudo_atoms:
                translated_residue_name = residue_name
            else:
                msg = f"""
                    The residue name {residue_name} is not defined for the molecule type {molecule_type}
                    at line {line_info.line_no} in file {line_info.file_name} the line was
                    {line_info.line}
                """
                exit_error(msg)

        residue_name_key = (chain_code, sequence_code)

        if residue_name == UNUSED and residue_name_lookup is None:
            msg = f"""
                no sequence was loaded and the assignment {assignments} while it has a sequence code {sequence_code}
                and chain code {chain_code} doesn't define a residue type, please provide a sequence in the input
                stream
            """
            exit_error(msg)
        else:

            if (
                residue_name_lookup is not None
                and residue_name_key in residue_name_lookup
            ):
                translated_residue_name = residue_name_lookup[residue_name_key]

        if (
            len(sequence) > 0
            and (residue_name_key not in residue_name_lookup)
            and residue_name_key[1] != UNUSED
        ):
            msg = f"""
                the chain code {chain_code} and sequence_code {sequence_code} from
                line {line_info.line_no} in file {line_info.file_name} were not found
                in the input sequence, the full line was

                {line_info.line}

                if you wish to input the peaks without validating against the input sequence use the
                --no-validate option of sparky import peaks
            """
            exit_error(msg)

        if residue_name_key[1] == UNUSED:
            translated_residue_name = UNUSED

        residue = Residue(
            chain_code=chain_code,
            sequence_code=sequence_code,
            residue_name=translated_residue_name,
        )
        assignment = AtomLabel(residue, atom_name)

        result.append(assignment)

        last_sequence_code = sequence_code
        last_residue_name = residue_name

    return result


def parse_peaks(
    lines, file_name, molecule_type, chain_code, sequence, allow_pseudo_atoms=False
):

    peaks = []

    in_data = False

    dimension_count = None

    for line_number, line in enumerate(lines, start=1):

        line = line.strip()

        line_info = LineInfo(file_name, line_number, line)

        if len(line) == 0:
            continue

        if line.startswith("#"):
            continue

        if line.strip()[: len("Assignment")] == "Assignment":
            if in_data:
                _exit_error_header_in_data(line_info)

            column_headers_to_indices = parse_header_to_columns(line, file_name)

            dimension_count = _count_dimensions(column_headers_to_indices)

            _exit_error_no_dimensions(dimension_count, line_info)

            in_data = True

            continue
        else:
            if not in_data:
                _exit_error_data_but_no_header(line_info)

            fields = line.split()

            column_count = len(fields)
            if column_count < (dimension_count + 1):
                _exit_error_not_enough_columns_in_data_row(
                    dimension_count, column_count, line_info
                )
            values = {}
            for column, index in column_headers_to_indices.items():
                if index < len(fields):
                    values[column] = fields[index]
                else:
                    values[column] = UNUSED

            assignmnents_column = column_headers_to_indices["Assignment"]
            raw_assignment = fields[assignmnents_column]

            assignments = parse_assignments(
                raw_assignment,
                chain_code,
                sequence,
                molecule_type,
                line_info,
                allow_pseudo_atoms=allow_pseudo_atoms,
            )

            shifts = [
                fields[column_headers_to_indices[f"w{index}"]]
                for index in range(1, dimension_count + 1)
            ]

            converted_shifts = []
            for shift in shifts:
                shift = float(shift) if is_float(shift) else shift
                converted_shifts.append(shift)

            _exit_error_if_shift_not_float(shifts, line_info)

            shifts = converted_shifts

            peak_fit_method = None
            volume = None
            if "Volume" in column_headers_to_indices:

                volume_index = column_headers_to_indices["Volume"]

                volume = float(fields[volume_index])

                line_fit_index = volume_index + 1
                if line_fit_index < len(fields):
                    if fields[line_fit_index] == "ga":
                        peak_fit_method = PeakFitMethod.GAUSSIAN
                        fields.remove("ga")

            height = (
                float(fields[column_headers_to_indices["Height"]])
                if "Height" in column_headers_to_indices
                else None
            )

            comment_fields = []
            line_widths = []
            for dimension in range(1, dimension_count + 1):
                line_width_column = f"lw{dimension}"
                if line_width_column in column_headers_to_indices:
                    line_width_column_index = column_headers_to_indices[
                        line_width_column
                    ]
                    if line_width_column_index < len(fields):
                        possible_line_width = fields[line_width_column_index]
                        if is_float(possible_line_width):
                            line_widths.append(float(possible_line_width))
                        else:
                            comment_fields.append(possible_line_width)
                            line_widths.append(None)
                    else:
                        line_widths.append(None)

            shift_data = [
                ShiftData(atom=atom, value=value, line_width=line_width)
                for atom, value, line_width in zip_longest(
                    assignments, shifts, line_widths, fillvalue=None
                )
            ]

            max_column = max(column_headers_to_indices.values())
            if len(fields) > max_column:
                comment_fields = [*comment_fields, *fields[max_column + 1 :]]

            comment = " ".join(comment_fields)

            peak = NewPeak(
                shifts=shift_data,
                peak_fit_method=peak_fit_method,
                height=height,
                volume=volume,
                comment=comment,
            )

            peaks.append(peak)

    return peaks


def parse_header_to_columns(header_line: str, file_name) -> Dict[str, int]:
    headings_to_columns = {}
    headings = header_line.split()

    if "Data" in headings:
        headings.remove("Data")

    while ("(hz)") in headings:
        headings.remove("(hz)")

    for i, heading in enumerate(headings):
        if heading in "Assignment Height Volume".split():
            headings_to_columns[heading] = i
            continue

        if heading.startswith("w") or heading.startswith("lw"):
            headings_to_columns[heading] = i
            continue

        # TODO add a warning function in the library
        msg = f"""
            WARNING: unexpected heading {heading} in the file {file_name}...
                     this heading will be ignored, please send the heading and first
                     few lines of this file to the developers of NEF-Pipelines if you
                     believe this is a valid sparky peaks file
        """

        print(msg, file=sys.stderr)

    assignment_column = headings_to_columns["Assignment"]
    if assignment_column != 0:
        msg = f"""
            The file {file_name} doesn't look like a sparky file the Assignment column should be the first column
            i got {assignment_column}
            header was
            {header_line}
        """
        exit(msg)

    return headings_to_columns


def _exit_error_if_shift_not_float(shifts, line_info):
    for i, shift in enumerate(shifts, start=1):
        if not is_float(shift):
            msg = f"""
                file {line_info.file_name} does not look like a sparky file
                for shift w{i} at line {line_info.line_no} with the value {shift} couldn't be converted to a float
                the full line was:
                {line_info.line}
            """
            exit_error(msg)


def _exit_error_data_but_no_header(line_info):
    msg = f"""
                    the file {line_info.file_name} doesn't look like a sparky file at line {line_info.line_no}
                    there appears to be a data line but no header was detected a head should look like
                    Assignment w1 w2 ... etc
                    the current line data is
                    {line_info.line}
                """
    exit_error(msg)


def _exit_error_not_enough_columns_in_data_row(
    dimension_count, column_count, line_info
):
    msg = f"""
            In sparky peaks file {line_info.file_name} at line {line_info.line_no}
            there were was not enough data [expected {1+ dimension_count} columns
            (Assignment + shifts * {dimension_count})]
            the line was:
            {line_info.line}
        """
    exit_error(msg)


def _exit_error_no_dimensions(dimension_count, line_info):
    if dimension_count < 1:
        msg = f"""
                    sparky peak file {line_info.file_name} doesn't appear to have enough columns [minimum 1]
                    at line {line_info.line_no}
                    the header was:
                    {line_info.line}
                """
        exit_error(msg)


def _count_dimensions(column_headers_to_indices):
    dimension_count = 0
    for name in [f"w{i}" for i in range(1, 20)]:
        if name in column_headers_to_indices:
            dimension_count += 1
        else:
            break
    return dimension_count


def _exit_error_header_in_data(line_info):
    msg = f"""
        bad sparky peak file {line_info.file_name} at line no {line_info.line_number} there appears to be a
        second header...
        the line was:
        {line_info.line}
                """
    exit_error(msg)



import string
import sys

import requests
from pynmrstar import Entry
from requests.exceptions import HTTPError
from tabulate import tabulate

from nef_pipelines.lib.util import exit_error, is_int
from nef_pipelines.transcoders.nmrstar.importers.project_shortcuts import (
    SHORTCUT_URLS,
    SHORTCUTS,
)
from nef_pipelines.transcoders.nmrstar.importers.sequence import pipe as sequence_pipe
from nef_pipelines.transcoders.nmrstar.importers.shifts import pipe as shift_pipe


def pipe(
    nef_entry,
    nmrstar_entry,
    chain_codes,
    no_chain_starts,
    no_chain_ends,
    use_author,
    stereo_mode,
    file_path,
):

    entry_id = nef_entry.entry_id
    nef_entry = sequence_pipe(
        nef_entry,
        chain_codes,
        [],
        no_chain_starts,
        no_chain_ends,
        entry_id,
        nmrstar_entry,
        file_path,
        use_author,
    )

    nef_entry = shift_pipe(
        nef_entry, [], None, nmrstar_entry, file_path, use_author, stereo_mode
    )

    return nef_entry


def _list_shortcuts_and_exit():
    print("The available shortcuts are:\n", file=sys.stderr)
    shortcuts = [
        [shortcut.lower(), entry, SHORTCUT_URLS[shortcut]]
        for shortcut, entry in SHORTCUTS.items()
    ]
    HEADERS = ["name", "entry id.", "url"]
    print(tabulate(shortcuts, headers=HEADERS), file=sys.stderr)
    print("\nexiting...", file=sys.stderr)
    sys.exit(0)


def _parse_text_to_star_or_none(possible_entry):
    nmrstar_entry = None
    if possible_entry and possible_entry.startswith(b"data_"):
        try:
            nmrstar_entry = Entry.from_string(possible_entry.decode("utf-8"))
        except Exception:
            pass
    return nmrstar_entry


def _get_bmrb_entry_from_web_or_none(
    urls, exit_on_error=True, verbose=False, timeout=10
):
    possible_entry = None

    have_data = False
    if urls:

        for i, url in enumerate(urls, start=1):
            error = False
            if verbose:
                print(f"{i}. trying to from download {url}", file=sys.stderr)
            try:
                try:
                    response = requests.get(url, timeout=timeout)
                    response.raise_for_status()
                    possible_entry = response.content
                except requests.exceptions.SSLError:
                    print(
                        f"NOTE: using weaker SSL key as long keys not supported by some mirrors [{url}]",
                        file=sys.stderr,
                    )
                    requests.packages.urllib3.disable_warnings()
                    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += (
                        "HIGH:!DH:!aNULL"
                    )
                    try:
                        requests.packages.urllib3.contrib.pyopenssl.DEFAULT_SSL_CIPHER_LIST += (
                            "HIGH:!DH:!aNULL"
                        )
                    except AttributeError:
                        # no pyopenssl support used / needed / available
                        pass
                    response = requests.get(url, verify=False, timeout=timeout)
                    response.raise_for_status()
                    possible_entry = response.content

            except HTTPError as http_err:
                if verbose:
                    msg = f"WARNING: while trying to download the url {url} there was an http error {http_err}"
                    print(msg, file=sys.stderr)
                    error = True

            except Exception as err:
                if verbose:
                    msg = f"WARNING: while trying to download the url {url} there was an error {err}"
                    print(msg, file=sys.stderr)
                    error = True

            if not error:
                have_data = True
                break

    if not have_data:
        if exit_on_error:
            urls = " /n".join(urls)
            msg = f"""\
            could not download the entry from

            {urls}
            """
            exit_error(msg)

    return possible_entry


def _get_path_as_url_or_none(file_path, url_templates):
    urls = []
    is_bmrb = False
    if file_path.startswith("https://"):
        urls = [
            file_path,
        ]
    elif file_path.startswith("bmr"):
        entry_number = file_path[3:]
        entry_check = entry_number.lstrip(string.digits)
        if len(entry_check) == 0:
            urls = [
                url_template.format(entry_number=entry_number)
                for url_template in url_templates
            ]
            is_bmrb = True
    elif is_int(file_path):
        entry_number = file_path
        urls = [
            url_template.format(entry_number=entry_number)
            for url_template in url_templates
        ]
        is_bmrb = True

    return urls, is_bmrb



import itertools
import string
import sys
from copy import copy
from dataclasses import dataclass, replace
from datetime import time
from enum import auto
from itertools import product
from pathlib import Path
from random import seed as set_random_seed
from random import shuffle
from typing import Dict, List

import typer
from ordered_set import OrderedSet
from pynmrstar import Saveframe
from strenum import KebabCaseStrEnum, LowercaseStrEnum

from nef_pipelines.lib.nef_lib import (
    SELECTORS_LOWER,
    UNUSED,
    SelectionType,
    loop_row_dict_iter,
    read_or_create_entry_exit_error_on_bad_file,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import atom_sort_key
from nef_pipelines.lib.structures import AtomLabel, Residue
from nef_pipelines.lib.util import (
    STDIN,
    chunks,
    exit_error,
    is_int,
    parse_comma_separated_options,
    strip_characters_left,
)
from nef_pipelines.tools.frames import frames_app

# TODO: add the ability to do inverted assignments with ^
# TODO add possibility of using complete instead of all... rather than having complete as an extra option?
RESIDUE_RANGES_HELP = """
    residue ranges to unassign, can be defined as comma separated pairs of ranges e.g 1-10,11-24
    or can also have a colon separated chain_code e.g A:1-10,B:14-32. Multiple selections are combined
    additively. Multiple copies of the option can be used to add extra values [default is all]
"""

SELECTOR_HELP = """
    limit changes to a a particular frame by a selector which can be a frame name or category
    note: wildcards [*] are allowed. Frames are selected by name and subsequently by category if the name
    doesn't match [-t /--selector-type allows you to force which selection type to use]. If no frame
    names or categories are provided all frames are unassigned.
"""

ATOM_LABEL_FIELDS = "chain_code sequence_code residue_name atom_name".split()


class Targets(KebabCaseStrEnum):
    CHAIN_CODE = auto()
    SEQUENCE_CODE = auto()
    RESIDUE_NAME = auto()
    # RESIDUE = auto()  #TODO add..
    ATOM_NAME = auto()
    ALL = auto()
    NOT_ATOM = auto()


ALL_COMPONENTS = {
    Targets.CHAIN_CODE,
    Targets.SEQUENCE_CODE,
    Targets.RESIDUE_NAME,
    Targets.ATOM_NAME,
}
NOT_ATOM_COMPONENTS = copy(ALL_COMPONENTS)
NOT_ATOM_COMPONENTS.remove(Targets.ATOM_NAME)

TARGET_LOOKUP = {
    str(Targets.CHAIN_CODE).replace("_", "-"): {
        Targets.CHAIN_CODE,
    },
    str(Targets.SEQUENCE_CODE).replace("_", "-"): {
        Targets.SEQUENCE_CODE,
    },
    str(Targets.RESIDUE_NAME).replace("_", "-"): {
        Targets.RESIDUE_NAME,
    },
    str(Targets.ATOM_NAME).replace("_", "-"): {
        Targets.ATOM_NAME,
    },
    str(Targets.ALL): ALL_COMPONENTS,
    str(Targets.NOT_ATOM).replace("_", "-"): NOT_ATOM_COMPONENTS,
    # str(Targets.RESIDUE): { #TODO add
    #     Targets.SEQUENCE_CODE,
    #     Targets.RESIDUE_NAME,
    # },
}

# TODO allow residue as an unassignment
# TARGETS_HELP = f"""
#     choose what to de-assign it can be a comma separated list or called multiple times. Possible values are:
#     {', '.join([e.value for e in Targets])}.
#     all, residue and not_atom are combinations, all is: chain_code + sequence_code + residue_name + atom_name and
#     residue is sequence_code + residue_name, not_atom is: chain_code + sequence_code + residue_name.
# """

# TODO: allow unique abbreviations
TARGETS_HELP = f"""
    choose what to de-assign it can be a comma separated list or called multiple times. Possible values are:
    {', '.join([e.value for e in Targets])}.
    all and not_atom are combinations, all is: chain_code + sequence_code + residue_name + atom_name and
    not_atom is: chain_code + sequence_code + residue_name. Complete makes assignments go back to completely empty
    assignments as opposed to pseudo atoms.
"""


# could do with an ordered mode which would sort assignments
# but would require a shift sorter
class SequenceMode(LowercaseStrEnum):
    UNUSED = auto()
    INDEX = auto()
    ORDERED = auto()
    RANDOM = auto()
    PRESERVE = auto()


# TODO: how do you decide whether its a selection within a chain or the file
SEQUENCE_MODE_HELP = f"""
    value to deassign the sequence code to [one of {', '.join([e for e in SequenceMode])}] UNUSED is .,
    INDEX is the index of the assignment in its chain, RANDOM is a random index for the assignment
    within its chain, ORDERED retains the order of the residues when deassigned to pseudo residues, PRESERVE deassigns
    residues to pseudo residues with the same number as the assigned residue any clashes are disambiguated by renaming
    existing pseudo residues.
"""

CHAIN_VALUE_HELP = """
    value to deassign chains to [note @- is the default nmr chain and . is no chain] if more than one value is specified
    [using multiple options or a comma separated list] they will be treated as old and new value pairs, with old values
    being reassigned to the new values. Current values which are not in the new values will be left untouched.
"""

RANDOM_SEED_HELP = """
    set the random seed to an integer, any value less than zero will set the seed from the current clock and
    note the value
"""
ANY = "*"

SELECTION_TYPE_HELP = f"""
    how to select frames to unassign, can be one of: {SELECTORS_LOWER}.
    Any will match on frame name first and then if there is no match attempt to match on category
"""

USE_RESIDUE_OFFSETS_HELP = """
if the experiment is a triple resonance experiment don't offset residues so that assignment from the previous frame
are labelled as CA-1 and CB-1 [compatability with CCPN]
"""
# TODO: add the ability to unassign a particular dimension
# TODO: add the ability specify the pseudo residue prefix


@dataclass
class ResidueRange:
    chain_code: str
    start: int
    end: int


COMPLETE_HELP = "if set completely unassign to undefined rather than a pseudo residue"


@frames_app.command()
def unassign(
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY, "-t", "--selector-type", help=SELECTION_TYPE_HELP
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    targets: List[Targets] = typer.Option(
        [Targets.NOT_ATOM], "-t", "--targets", help=TARGETS_HELP
    ),
    chain_values: List[str] = typer.Option(
        [], metavar="<CHAIN> | <OLD-CHAIN>: <NEW-CHAIN>...", help=CHAIN_VALUE_HELP
    ),
    sequence_mode: SequenceMode = typer.Option(
        SequenceMode.INDEX, metavar="<MODE>", help=SEQUENCE_MODE_HELP
    ),
    random_seed: int = typer.Option(42, metavar="<RANDOM-SEED>", help=RANDOM_SEED_HELP),
    no_residue_offsets: bool = typer.Option(
        False, "--no-residue-offsets", help=USE_RESIDUE_OFFSETS_HELP
    ),
    raw_residue_ranges: List[str] = typer.Option(
        None,
        "--residue-ranges",
        help=RESIDUE_RANGES_HELP,
    ),
    complete: bool = typer.Option(False, help=COMPLETE_HELP),
    frame_selectors: List[str] = typer.Argument(
        None,
        help=SELECTOR_HELP,
    ),
):
    """- unassign frames in the current input [alpha] either to pseudo residues or completely [option --complete]"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    raw_residue_ranges = parse_comma_separated_options(raw_residue_ranges)

    residue_ranges = _build_residue_ranges(raw_residue_ranges)

    use_residue_offsets = not no_residue_offsets

    if random_seed < 0:
        seed = int(time.time())
        set_random_seed(seed)
        print(f"SEED is set to: {seed}", file=sys.stderr)
    else:
        set_random_seed(random_seed)

    chain_values = parse_comma_separated_options(chain_values)
    num_chain_values = len(chain_values)
    if num_chain_values == 0:
        chain_mappings = {ANY: "@-"}
    elif num_chain_values > 1 and num_chain_values % 2 != 0:
        msg = f"""
            if there are more than two new chain_code values specified they must come in pairs i got
            {num_chain_values}, the values were {', '.join(chain_values)}
        """
        exit_error(msg)

    elif num_chain_values == 1:
        chain_mappings = {ANY: chain_values[0]}
    else:
        chain_mappings = {old: new for old, new in chunks(chain_values, 2)}

    raw_targets = parse_comma_separated_options(targets)

    targets = _build_targets(raw_targets)

    if Targets.CHAIN_CODE in targets and len(chain_mappings) == 0:
        chain_mappings_msg = ", ".join(
            [f"{first} -> {second}" for first, second in chain_mappings.items()]
        )

        msg = f"""\
            you can't have chain mappings and not target chains, you might want to add CHAIN_CODE to --targets!
            targets were: {', '.join([str(target) for target in targets])}
            chain mappings were: {chain_mappings_msg}
        """

        exit_error(msg)

    frame_selectors = {
        (frame.category, frame.name)
        for frame in select_frames(entry, frame_selectors, selector_type)
    }

    entry = pipe(
        entry,
        frame_selectors,
        targets,
        sequence_mode,
        chain_mappings,
        residue_ranges,
        use_residue_offsets,
        complete,
    )

    print(entry)


# noinspection PyUnusedLocal


# todo move frame selectors to non pipe part
def pipe(
    entry,
    frame_selectors,
    targets,
    sequence_mode,
    chain_mappings,
    residue_ranges,
    use_residue_offsets,
    complete,
):

    all_assignments = OrderedSet()

    target_frames = [
        frame for frame in entry if (frame.category, frame.name) in frame_selectors
    ]

    # this is a bit messy it uses the entry for storing data...
    for frame in target_frames:
        if use_residue_offsets:
            _offset_sequence_codes_in_triple(frame, targets)

    for frame in target_frames:
        all_assignments.update(_build_current_assignments_for_frame(frame))

    assignments_to_remove = _select_assignments_to_remove_by_residue_ranges(
        all_assignments, residue_ranges, use_residue_offsets
    )

    if sequence_mode == SequenceMode.ORDERED:
        assignments_to_remove = sorted(assignments_to_remove, key=atom_sort_key)

    assignment_map = _map_assignments(
        assignments_to_remove, targets, sequence_mode, chain_mappings, complete
    )

    for frame in target_frames:
        for loop in frame:
            _reassign_loop(loop, assignment_map)

    for frame in target_frames:
        if use_residue_offsets:
            _unoffset_sequence_codes_in_triple(frame, targets)

    return entry


def _build_residue_ranges(raw_residue_ranges: List[str]) -> List[ResidueRange]:

    results = []
    for elem in raw_residue_ranges:
        if ":" in elem:
            chain_code, range_str = elem.split(":")
        else:
            chain_code = ANY
            range_str = elem

        range_fields = range_str.split("-")

        if len(range_fields) != 2:
            msg = f"""
                residue ranges must be in the form <START>-<END> [2 elements], i got {elem} [{len(range_fields)}]
            """
            exit_error(msg)

        for i, field in enumerate(range_fields):
            if not is_int(field):
                msg = f"""
                    residue ranges must be integers, i got {field} for element {i} in {elem}
                """
                exit_error(msg)

            range_fields[i] = int(field)

        start, end = range_fields

        if end >= start:
            results.append(ResidueRange(chain_code, start, end))

    if not results:
        results.append(ResidueRange(ANY, -sys.maxsize, sys.maxsize))
    return results


def _select_assignments_to_remove_by_residue_ranges(
    current_assignments, residue_ranges, use_residue_offsets
):
    assignments_to_remove = OrderedSet()

    current_assignments = OrderedSet(current_assignments)
    for assignment in current_assignments:
        for residue_range in residue_ranges:
            # TODO: add a specific function to check residue assignment status
            if (
                residue_range.chain_code == ANY
                or residue_range.chain_code == assignment.residue.chain_code
            ):
                if (
                    is_int(assignment.residue.sequence_code)
                    and residue_range.start
                    <= assignment.residue.sequence_code
                    <= residue_range.end
                ):
                    assignments_to_remove.add(assignment)
                    break
                elif (
                    assignment.residue.chain_code == "@-"
                    and isinstance(assignment.residue.sequence_code, str)
                    and assignment.residue.sequence_code[0] == "@"
                    and is_int(assignment.residue.sequence_code[1:])
                ):
                    break  # already fully unassigned ignore
                elif (
                    residue_range.start == -sys.maxsize
                    and residue_range.end == sys.maxsize
                    and assignment.residue.sequence_code[0] in "@#"
                ):
                    assignments_to_remove.add(assignment)
                    break

    return assignments_to_remove


def _reassign_loop(loop, assignment_map):
    tag_indices = {tag: index for index, tag in enumerate(loop.tags)}
    for row_index, row in enumerate(loop_row_dict_iter(loop)):
        assignments_by_index = {}
        for index, prefix in product(range(1, 16), ATOM_LABEL_FIELDS):

            if prefix in row:
                key = prefix
            else:
                key = f"{prefix}_{index}"

            if key not in row:
                break

            assignments_by_index.setdefault(index, {})[prefix] = row[key]

        for column_index, assignment in enumerate(
            assignments_by_index.values(), start=1
        ):

            old_label = _dict_to_label(assignment)

            if old_label is not None and old_label in assignment_map:

                new_label = assignment_map[old_label]

                for atom_label_field in ATOM_LABEL_FIELDS:
                    key = None
                    if atom_label_field in tag_indices:
                        key = atom_label_field
                    if not key:
                        possible_key = f"{atom_label_field}_{column_index}"
                        if possible_key in tag_indices:
                            key = possible_key

                    if key:
                        tag_index = tag_indices[key]

                        if atom_label_field == "chain_code":
                            new_value = new_label.residue.chain_code
                        elif atom_label_field == "sequence_code":
                            if new_label.residue.offset == 0:
                                new_value = new_label.residue.sequence_code
                            else:
                                if new_label.residue.sequence_code != UNUSED:
                                    new_value = f"{new_label.residue.sequence_code}{new_label.residue.offset}"
                                else:
                                    new_value = new_label.residue.sequence_code

                        elif atom_label_field == "residue_name":
                            new_value = new_label.residue.residue_name
                        elif atom_label_field == "atom_name":
                            new_value = new_label.atom_name

                        loop.data[row_index][tag_index] = new_value


def _build_current_assignments_for_frame(frame):

    result = OrderedSet()

    for loop in frame:
        for i, row in enumerate(loop_row_dict_iter(loop)):
            assignments_by_index = {}
            for index, prefix in product(range(1, 16), ATOM_LABEL_FIELDS):

                if prefix in row:
                    key = prefix
                else:
                    key = f"{prefix}_{index}"

                    if key not in row:
                        break

                assignments_by_index.setdefault(index, {})[prefix] = row[key]

            for assignment in assignments_by_index.values():
                label = _dict_to_label(assignment)

                if label:
                    result.add(label)

    return result


def _build_targets(raw_targets: List[Targets]) -> List[Targets]:
    result = set()

    for raw_target in raw_targets:
        result.update(TARGET_LOOKUP[raw_target])

    return result


def _dict_to_label(assignment):
    result = None
    if "atom_name" in assignment:
        atom_name = assignment["atom_name"]
        residue_values = {
            key: assignment[key]
            for key in assignment
            if key != "atom_name" and key in ATOM_LABEL_FIELDS
        }
        if "sequence_code" in residue_values:
            sequence_code = residue_values["sequence_code"]
            if is_int(sequence_code):
                residue_values["sequence_code"] = int(sequence_code)
            elif sequence_code_offset := _get_sequence_code_and_offset_or_none(
                sequence_code
            ):
                sequence_code, offset = sequence_code_offset
                residue_values["sequence_code"] = sequence_code
                residue_values["offset"] = offset
            else:
                residue_values["sequence_code"] = sequence_code

        residue = Residue(**residue_values)
        result = AtomLabel(residue, atom_name)
    return result


def _get_sequence_code_and_offset_or_none(sequence_code):

    result = None
    sequence_code_fields = sequence_code.split("-") if "-" in str(sequence_code) else []
    if len(sequence_code_fields) == 2:
        sequence_code, offset = sequence_code_fields
        if is_int(sequence_code) and is_int(offset):
            sequence_code = int(sequence_code)
            offset = -int(offset)

            result = sequence_code, offset
    return result


TRIPLE_RESONANCE_CLASSIFICATION = [
    "H[N[co[CA]]]",
    "H[N[CA[CB]]]",
    "H[N[co[CA[CB]]]]",
    "H[N[CA]]",
    "H[N[ca[CO]]]",
    "H[N[CO]]",
]


def _is_triple_resonance_frame(frame: Saveframe):
    result = False
    if frame.category == "nef_nmr_spectrum":
        experiment_classification = frame.get_tag("experiment_classification")
        experiment_classification = (
            experiment_classification[0] if experiment_classification else None
        )
        if experiment_classification in TRIPLE_RESONANCE_CLASSIFICATION:
            result = True
    return result


def _map_assignments(
    assignments: List[AtomLabel], targets, sequence_mode, chain_mappings, complete
) -> Dict[AtomLabel, AtomLabel]:

    assignment_map = {}
    if Targets.CHAIN_CODE in targets:
        for assignment in assignments:

            if complete:
                new_chain_code = UNUSED
            else:
                if sequence_mode == SequenceMode.UNUSED:

                    new_chain_code = "."
                elif len(chain_mappings) == 1 and list(chain_mappings.keys())[0] == ANY:
                    new_chain_code = list(chain_mappings.values())[0]
                else:
                    target_chain_code = assignment.residue.chain_code
                    new_chain_code = target_chain_code
                    if assignment.residue.chain_code in chain_mappings:
                        new_chain_code = chain_mappings[target_chain_code]

            new_residue = replace(assignment.residue, chain_code=new_chain_code)
            new_assignment = replace(assignment, residue=new_residue)

            assignment_map[assignment] = new_assignment

    if Targets.RESIDUE_NAME in targets:
        for assignment, new_assignment in assignment_map.items():

            new_residue = replace(new_assignment.residue, residue_name=UNUSED)
            new_assignment = replace(new_assignment, residue=new_residue)

            assignment_map[assignment] = new_assignment

    if Targets.ATOM_NAME in targets:
        for assignment, new_assignment in assignment_map.items():

            new_assignment = replace(new_assignment, atom_name=UNUSED)

            assignment_map[assignment] = new_assignment

    if Targets.SEQUENCE_CODE in targets:

        if complete:
            for assignment, new_assignment in assignment_map.items():

                new_residue = replace(new_assignment.residue, sequence_code=UNUSED)
                new_assignment = replace(new_assignment, residue=new_residue)

                assignment_map[assignment] = new_assignment
        else:
            new_assignment_by_chain = {}
            for assignment, new_assignment in assignment_map.items():

                chain_code = new_assignment.residue.chain_code
                new_assignment_by_chain.setdefault(chain_code, []).append(
                    (assignment, new_assignment)
                )

            sequence_code_map = {}

            if len(new_assignment_by_chain) > 1:
                msg = """
                    !WARNING! there is more than one chain present, this code hasn't been tested with multiple chains...
                """
                print(msg, file=sys.stderr)

            for chain, assignments in new_assignment_by_chain.items():
                old_sequence_codes = OrderedSet()
                for old_assignment, new_assignment in assignments:

                    sequence_code = str(new_assignment.residue.sequence_code)

                    old_sequence_codes.add(sequence_code)

                if sequence_mode == SequenceMode.UNUSED:
                    for old_sequence_code in old_sequence_codes:
                        sequence_code_map[old_sequence_code] = UNUSED

                elif sequence_mode in (
                    SequenceMode.INDEX,
                    SequenceMode.RANDOM,
                    SequenceMode.ORDERED,
                ):
                    # pull out unassigned sequence_codes
                    used_pseudo_residue_sequence_codes = [
                        sequence_code
                        for sequence_code in old_sequence_codes
                        if sequence_code[0] not in string.digits
                    ]
                    old_sequence_codes = [
                        sequence_code
                        for sequence_code in old_sequence_codes
                        if sequence_code[0] in string.digits
                    ]

                    new_sequence_code_counter = itertools.count(1)
                    new_sequence_codes = []
                    for _ in old_sequence_codes:

                        putative_sequence_code = next(new_sequence_code_counter)
                        while (
                            f"@{putative_sequence_code}"
                            in used_pseudo_residue_sequence_codes
                        ):
                            putative_sequence_code = next(new_sequence_code_counter)

                        new_sequence_codes.append(f"@{putative_sequence_code}")

                    if sequence_mode == SequenceMode.RANDOM:
                        shuffle(new_sequence_codes)

                    for i, old_sequence_code in enumerate(old_sequence_codes):
                        sequence_code_map[old_sequence_code] = new_sequence_codes[i]

                    for sequence_code in used_pseudo_residue_sequence_codes:
                        sequence_code_map[sequence_code] = sequence_code

                elif sequence_mode == SequenceMode.PRESERVE:

                    # TODO: note there is a subtle difference here to above can this be removed
                    used_pseudo_residue_sequence_codes = [
                        sequence_code
                        for sequence_code in old_sequence_codes
                        if not is_int(sequence_code)
                    ]
                    old_sequence_codes = [
                        sequence_code
                        for sequence_code in old_sequence_codes
                        if is_int(sequence_code)
                    ]

                    sequence_code_map = {}

                    for sequence_code in old_sequence_codes:
                        sequence_code_map[sequence_code] = f"@{sequence_code}"

                    possible_new_sequence_codes = {}
                    for used_sequence_code in used_pseudo_residue_sequence_codes:
                        if used_sequence_code == UNUSED:
                            possible_new_sequence_codes[UNUSED] = UNUSED
                        else:
                            for i in itertools.chain(
                                [
                                    "",
                                ],
                                itertools.count(1),
                            ):

                                prefix, sequence_code = strip_characters_left(
                                    used_sequence_code, string.ascii_letters + "@#"
                                )

                                new_sequence_code = f"{prefix}PR_{i}{sequence_code}"
                                if new_sequence_code in sequence_code_map:
                                    continue

                                possible_new_sequence_codes[used_sequence_code] = (
                                    new_sequence_code
                                )
                                break

                    sequence_code_map.update(possible_new_sequence_codes)

            for assignment, new_assignment in assignment_map.items():
                old_sequence_code = str(new_assignment.residue.sequence_code)
                offset = None
                if "-" in old_sequence_code:
                    fields = old_sequence_code.split("-")
                    if is_int(fields[-1]):
                        offset = fields[-1]
                        old_sequence_code = "-".join(fields[:-1])

                new_sequence_code = sequence_code_map[old_sequence_code]
                if sequence_mode == SequenceMode.UNUSED:
                    new_sequence_code = UNUSED
                elif offset is not None:
                    new_sequence_code = f"{new_sequence_code}-{offset}"

                updated_residue = replace(
                    new_assignment.residue, sequence_code=new_sequence_code
                )

                if updated_residue.sequence_code == UNUSED:
                    updated_residue = replace(updated_residue, chain_code=UNUSED)

                updated_atom_label = replace(new_assignment, residue=updated_residue)

                assignment_map[assignment] = updated_atom_label

    return assignment_map


def _offset_sequence_codes_in_triple(frame, targets):
    if _is_triple_resonance_frame(frame) and Targets.SEQUENCE_CODE in targets:
        peak_loop = frame.get_loop("nef_peak")

        sequence_codes = set()
        for row in loop_row_dict_iter(peak_loop):
            for i in range(1, 16):
                key = f"{'sequence_code'}_{i}"
                if key in row:
                    sequence_code = row[key]
                    if is_int(sequence_code):
                        sequence_codes.add(int(sequence_code))

            if sequence_codes:
                max_sequence_code = max(sequence_codes)
                max_sequence_code_m1 = max_sequence_code - 1

                for i in range(1, 16):
                    key = f"{'sequence_code'}_{i}"
                    if key in row:
                        sequence_code = row[key]
                        if is_int(sequence_code):
                            sequence_code = int(sequence_code)
                            if sequence_code == max_sequence_code_m1:
                                sequence_code = f"{max_sequence_code}-1"
                                row[key] = sequence_code


def _unoffset_sequence_codes_in_triple(frame, targets):
    if _is_triple_resonance_frame(frame) and Targets.SEQUENCE_CODE in targets:
        peak_loop = frame.get_loop("nef_peak")

        for row in loop_row_dict_iter(peak_loop):
            for i in range(1, 16):
                key = f"{'sequence_code'}_{i}"
                if key in row:
                    sequence_code = row[key]
                    is_assigned_residue = str(sequence_code)[0] != "@"
                    if is_assigned_residue:
                        if sequence_code_and_offset := _get_sequence_code_and_offset_or_none(
                            sequence_code
                        ):
                            sequence_code, offset = sequence_code_and_offset
                            sequence_code = sequence_code + offset
                            row[key] = sequence_code



import string
from collections import OrderedDict
from typing import Dict, Iterator, List, Tuple, Union

from nef_pipelines.lib.isotope_lib import ATOM_TO_ISOTOPE
from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.sequence_lib import get_residue_name_from_lookup
from nef_pipelines.lib.structures import (
    AtomLabel,
    DimensionInfo,
    LineInfo,
    NewPeak,
    Residue,
    SequenceResidue,
    ShiftData,
)
from nef_pipelines.lib.util import (
    exit_error,
    is_float,
    is_int,
    strip_characters_right,
    strip_line_comment,
)

XEASY_PEAK_MAGIC = "FORMAT xeasy3D"


def line_info_iter(lines: Iterator[str], source="unknown") -> Iterator[LineInfo]:
    for line_no, line in enumerate(lines, start=1):
        line = line.strip()

        if len(line) == 0:
            continue

        yield LineInfo(source, line_no, line)


def parse_sequence(lines: Iterator[str], source="unknown") -> List[SequenceResidue]:
    result = []
    for line_info in line_info_iter(lines, source):

        if line_info.line_no == 1 and line_info.line.startswith("#"):
            continue

        fields = line_info.line.split()

        exit_if_wrong_field_count_sequence(fields, line_info)

        residue_name, chain_seq_code = fields

        if residue_name == "SS":
            msg = f"""
                Original xeasy sequences are not supported only those used by flya
                at line {line_info.line_no} in file {line_info.file_name} if found a residue type SS
                where as it must be a standard 3 letter amino acid code, the whole line was

                {line_info.line}
            """
            exit_error(msg)

        chain_code, sequence_code = strip_characters_right(
            chain_seq_code, string.digits
        )

        sequence_residue = SequenceResidue(chain_code, sequence_code, residue_name)

        result.append(sequence_residue)

    return result


# A peak list contains an entry for each peak. The fields describing a peak are listed in Table 3.4.1.A. In a
# 2D spectrum the fields for the w3 and w4 dimension and in a 3D spectrum those for w4 are not used. The
# peak numbers in a peaklist must be unique but not necessarily continuous.
#
# Table 3.4.1.A Peak Fields
#
# Dim.    Fields          Description
#         peak number     unique number identifying the peak                                 1
#         colour          colour in the range [1,6] used for displaying the peaks            2
#         volume          volume of the peak                                                 3
#         volume error    volume error in percent                                            4
#         integration method                                                                 5
#                         method used for the integration: d, r, e, m, a, -
#         comment         user defined comment                                               6
#         possible ass.   data structure containing possible assignments                     7
#         shift           folded w1 position in ppm                                          8
# w1      fold            number of times peak is folded in w1                               9
#         atom number     assignment in w1: reference into the atom list                     10
#         shift           folded w2 position in ppm                                          11
# w2      fold            number of times peak is folded in w2                               12
#         atom number     assignment in w2: reference into the atom list                     13
#         shift           folded w3 position in ppm                                          14
# w3      fold            number of times peak is folded in w3                               15
#         atom number     assignment in w3: reference into the atom list                     16
#         shift           folded w4 position in ppm                                          17
# w4      fold            number of times peak is folded in w4                               18
#         atom number     assignment in w4: reference into the atom list                     19
# The peak list is stored in a peak list file with extension .peaks. It can be read by the program ASNO (P.
# Güntert et al., 1993) which uses a peak list, an atom list and selected structure coordinate files to generate
# possible assignments for NOESY cross peaks which can be loaded into XEASY. The program CALIBA
# (P. Güntert et al., J. Mol. Biol. (1991) 217, 517-530) translates the peak lists containing integrated peaks
# into distance constraints which can be used by the program DIANA (P. Güntert et al., J. Mol. Biol. (1991)
# 217, 517-530). An example of the first few lines of a two dimensional peak list is given below:
# # Number of dimensions 2
#    11  7.289  10.169 1 ?          2.048e+03  0.00e+00 -   0  126  128  0
#        #  first peak
#    12  7.119   9.413 1 ?          1.280e+02  0.00e+00 -   0  517  506  0
#    3   7.106   7.497 1 ?          4.096e+03  0.00e+00 -   0  129  130  0
#    4   7.228   7.411 1 ?          4.096e+03  0.00e+00 -   0  131  127  0
#    5   7.106   7.411 1 ?          5.120e+02  0.00e+00 -   0  327  328  0
#    6   6.838   7.094 1 ?          8.192e+03  0.00e+00 -   0  489  488  0
# The number on the first line after the hash "#" indicates the dimensionality of the peak list. Subsequent
# lines, starting with a number, contain the fields for one peak. Additional lines starting with a hash "#" are
# comments for the peak on the line above. The first field for each peak is the peak number, followed by:
# the unfolded chemical shift coordinates in ppm in w1 and w2 (more numbers are listed for higher
# dimensional spectra), the color code (a number from 1 to 6), the user defined type of the spectrum where
# the peak is observed, the peak volume, the uncertainty of the volume in percent, the integration method
# ("d" for Denk integration, "r" for rectangular integration, "e" for elliptical integration, "m" for maximum
# integration, "a" for automatic integration, "-" for not integrated), an unused number, the assignments in
# w1 and w2 is given by the two following atom numbers (more numbers are listed for higher dimensional
# spectra), the last number is not used. The commands to load or write peak lists are "Load peaklist [lp]"
# and "Write peaklist [wp]".

test_peaks = """
# Number of dimensions 3
#FORMAT xeasy3D
#INAME 1 HN
#INAME 2 H
#INAME 3 N
#SPECTRUM N15TOCSY HN H N
      1    8.731    8.723  115.265 1 U   5.500000E+00  0.000000E+00 e 0 H.A2     H.A2     N.A2 #MAP    403
      2    8.731    4.610  115.275 1 U   5.500000E+00  0.000000E+00 e 0 -         -         -
      3    8.732    4.192  115.254 1 U   5.500000E+00  0.000000E+00 e 0 H.A2     HA2.A2   N.A2 #MAP    404
      3    8.732    4.192  115.254 1 U   5.500000E+00  0.000000E+00 e 0 H.A2     HA3.A2   N.A2 #MAP    405
"""
# ser  shift1   shift2 shift3  colour  ??? vol           vol-error    int-method  a-number-or-folding  ass-1     ass-2     ass-3 comment      # noqa: E501 - this is not part of the file
# 1    8.731    8.723  115.265 1       U   5.500000E+00  0.000000E+00 e           0                    H.A82     H.A82     N.A82 #MAP    403  # noqa: E501 - this is not part of the file


def _parse_xeasy_assignment(
    assignment: str,
    residue_lookup: Dict[Tuple[Union[str, int], str], str],
    line_info: LineInfo,
) -> AtomLabel:

    if is_int(assignment):
        exit_error(f"pure xeasy assignments not currently supported, {line_info.line}")

    fields = assignment.split(".")
    if len(fields) == 1 and fields[0] == "-":
        residue = SequenceResidue(UNUSED, UNUSED, UNUSED)
        atom = AtomLabel(residue, UNUSED)

    elif len(fields) not in (1, 2):
        msg = f"""
            an xeasy assignment should either be an integer or a string of the for <ATOM>.<CHAIN_CODE><SEQUENCE_CODE>
            for example 10 or HA2.A82. I got {assignment} at {line_info.line_no} in file {line_info.file_name}
            the whole line was

            {line_info.line}

        """
        exit_error(msg)
    else:
        atom_name, chain_residue = fields
        chain_code, sequence_code = strip_characters_right(chain_residue, string.digits)

        residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_lookup
        )
        _exit_if_no_residue_name_in_sequence(
            residue_name, chain_code, sequence_code, line_info
        )

        residue = SequenceResidue(chain_code, sequence_code, residue_name)
        atom = AtomLabel(residue, atom_name)

    return atom


def _exit_if_no_residue_name_in_sequence(
    residue_name, chain_code, sequence_code, line_info
):

    if residue_name == UNUSED:
        msg = f"""
                in the file {line_info.file_name} at line {line_info.line_no} the sequence of the read molecules
                doesn't define a residue for the chain code {chain_code} and the sequence code {sequence_code},
                did you read a sequence (or the correct one)
                the line was:

                {line_info.line}
            """
        exit_error(msg)


def parse_peaks(
    lines: Iterator[str], source: str, residue_name_lookup: Dict[Union[str, int], str]
) -> Tuple[str, List[DimensionInfo], List[NewPeak]]:
    have_magic = False
    number_dimensions = None
    experiment_type = None
    dimensions = OrderedDict()

    peaks = []

    for line_info in line_info_iter(lines, source):

        shifts: list[float] = []
        if line_info.line.startswith("#"):

            stripped_line = line_info.line.lstrip("#").lstrip()

            if stripped_line.startswith(XEASY_PEAK_MAGIC):
                have_magic = True
                continue

            if stripped_line.startswith("Number of dimensions"):
                fields = stripped_line.split()
                number_dimensions = fields[-1]
                check_number_dimensions_or_exit(line_info, number_dimensions)
                number_dimensions = int(number_dimensions)
                continue

            if stripped_line.startswith("INAME"):

                fields = stripped_line.split()

                _check_iname_fields_length_or_exit(fields, line_info)

                dimension_number = fields[1]
                dimension_code = fields[2]

                check_dimension_number_is_integer_or_exit(dimension_number, line_info)

                dimension_number = int(dimension_number)
                dimensions[dimension_number - 1] = dimension_code
                continue

            if stripped_line.startswith("SPECTRUM"):
                fields = stripped_line.split()
                if len(fields) > 1:
                    experiment_type = fields[1]

        else:
            _exit_if_no_magic_before_data(have_magic, line_info)
            line, comment = strip_line_comment(line_info.line)

            fields = line.split()

            serial = fields[0]
            _check_serial_is_integer_or_exit(serial, line_info)
            serial = int(serial)
            current_column = 1

            for shift_column_number in dimensions:
                target_column = current_column + shift_column_number
                shift = fields[target_column]
                _check_shift_is_float_or_exit(shift, target_column, line_info)
                shifts.append(float(shift))
            current_column += number_dimensions

            # ignore peak color and U
            current_column += 2

            volume = fields[current_column]
            _check_volume_is_float_or_exit(volume, current_column, line_info)
            volume = float(volume)
            current_column += 1

            volume_error = fields[current_column]
            _check_volume_is_percentage_or_exit(volume_error, current_column, line_info)
            volume_error = float(volume_error)
            volume_error = volume * volume_error / 100.0
            current_column += 1

            # ignore the peak integration method
            current_column += 2

            assignments = []

            for assignment_column_number in dimensions:
                target_column = current_column + assignment_column_number
                assignment = _parse_xeasy_assignment(
                    fields[target_column], residue_name_lookup, line_info
                )
                assignments.append(assignment)

            shift_data = [
                ShiftData(assignment, shift)
                for assignment, shift in zip(assignments, shifts)
            ]
            peak = NewPeak(
                shifts=shift_data,
                id=serial,
                volume=volume,
                volume_uncertainty=volume_error,
                comment=comment,
            )

            peaks.append(peak)

    dimension_names = [
        dimensions[dimension_index] for dimension_index in sorted(dimensions)
    ]
    axis_codes = [
        ATOM_TO_ISOTOPE[dimension_name[0]] for dimension_name in dimension_names
    ]
    dimension_info = [
        DimensionInfo(axis_code, dimension_name)
        for axis_code, dimension_name in zip(axis_codes, dimension_names)
    ]

    return experiment_type, dimension_info, peaks


# 3.4.2 Atom List
#
# The atom list contains the names and frequencies of resonances. They define the possible assignments for
# each dimension of a peak.
#
# Table 3.4.2.A - Atom Fields
#
# Fields              Description
#
# atom number         unique number identifying the atom
# shift               mean chemical shift in ppm
# shift error         deviation of the assigned peaks from the mean value
# name                atom name
# fragment number     number of the fragment to which the atom belongs
# lineshapes          data structure containing the reference lineshapes
#
#
# The fields listed in Table 3.4.2.A constitute an atom entry. They can be modified in the peak editing
# window. The atom numbers, used to reference the atoms, must be unique but not necessarily continuous.
# The number -9999 is reserved to denote invalid entries. The average chemical shift and the shift error can
# be calculated from the assigned peaks. The command is "Average chem. shift [ac]".
#
# If the chemical shift is not defined it is set to the value 999.000.
#
# A new atom list is generated each time when a fragment list is loaded. For each fragment the
# corresponding atoms are looked up in the fragment library file and added to the list. New atom entries are
# added to the atom list if a non existing atom is used with the "Assign peak [ap]" command, if the fragment
# type is changed in the peak editing window, or with the "Add new fragment [af]" command.
#
# The atom list file has the extension ".prot" originating from the old EASY format. The following line is
# taken from such a file:
#
# 32   4.370  0.004   HA  2
#
# The first number is the atom number, followed by its mean chemical shift and the deviation from the
# mean value. The atom name and the fragment number follow. The commands to read or write an atom list
# are "Load atoms (chem. shift) [lc]" and "Write atoms (chem. shift) [wc]".


def parse_shifts(
    lines: Iterator[str], source: str, residue_lookup: Dict[Union[str, int], str]
):
    shifts = []
    for line_no, line in enumerate(lines, start=1):

        line_info = LineInfo(source, line_no, line.strip())

        if len(line_info.line) == 0:
            continue

        fields = line_info.line.split()

        _check_number_fields_correct_or_exit(fields, line_info)

        column = 1
        serial = fields[column - 1]
        _check_serial_is_integer_or_exit(serial, line_info)

        column += 1
        shift = fields[column - 1]
        _check_fields_is_float_or_exit(shift, column, line_info)

        column += 1
        shift_error = fields[column - 1]
        _check_fields_is_float_or_exit(shift_error, column, line_info)

        column += 1
        atom_name = fields[column - 1]

        column += 1
        chain_residue = fields[column - 1]
        chain_code, sequence_code = strip_characters_right(chain_residue, string.digits)

        residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_lookup
        )
        _exit_if_no_residue_name_in_sequence(
            residue_name, chain_code, sequence_code, line_info
        )

        residue = Residue(chain_code, sequence_code, residue_name)
        atom = AtomLabel(residue, atom_name)

        shift = ShiftData(atom=atom, value=shift, value_uncertainty=shift_error)

        shifts.append(shift)

    return shifts


def _check_fields_is_float_or_exit(value, column, line_info):
    if not is_float(value):
        msg = f"""
                column {column} of {line_info.file_name} at line {line_info.line_no} should be a floating point
                 number I got {value}, the line was

                {line_info.line}
            """
        exit_error(msg)


def _check_number_fields_correct_or_exit(fields, line_info):

    num_fields = len(fields)
    if num_fields != 5:
        msg = f"""
                in the file {line_info.file_name} at line {line_info.line_no} I expected 5 fields but got {num_fields}
                the line was

                {line_info}

                i expected something like

                32 4.370 0.004 HA A2
            """
        exit_error(msg)


def _exit_if_no_magic_before_data(have_magic, line_info):
    if not have_magic:
        msg = f"""
                    Xeasy files should have the line #{XEASY_PEAK_MAGIC} before any data is read
                    at line {line_info.line_no} in file {line_info.file_name} i appear to have data
                    before this occured. The line was

                    {line_info.line}
                """
        exit_error(msg)


def _check_volume_is_percentage_or_exit(volume_error, column_number, line_info):
    msg = None
    if not is_float(volume_error):
        msg = f"""
            the volume error column (column {column_number}) should be a floating point number
            i got {volume_error} in the file {line_info.file_name} at line {line_info.line_no} the value of the
            line was

            {line_info.line}
        """
    volume_error = float(volume_error)
    if not (100.0 >= volume_error >= 0.0):
        msg = f"""
            the volume error column (column {column_number}) should be a nunber between 0.0 and 100.0 i got
            {volume_error} in the file {line_info.file_name} at line {line_info.line_no} the value of the line was

            {line_info.line}
        """

    if msg:
        exit_error(msg)


def _check_volume_is_float_or_exit(volume, column_number, line_info):
    if not is_float(volume):
        msg = f"""
            the volume column (column {column_number}) should be a floating point number i got {volume}
            in the file {line_info.file_name} at line {line_info.line_no} the value of the line was

            {line_info.line}
        """
        exit_error(msg)


def _check_shift_is_float_or_exit(shift, column_number, line_info):
    if not is_float(shift):
        msg = f"""
                chemical shift fields should be floating point numbers i got {shift} at column number {column_number}
                in the file {line_info.file_name} at line {line_info.line_no} the value of the line was

                {line_info.line}
                    """
        exit_error(msg)


def _check_serial_is_integer_or_exit(serial, line_info):
    if not is_int(serial):
        msg = f"""
            the first field of a xeasy peak or shift  list should be an integer serial number i got: {serial}
            in file {line_info.file_name} at line {line_info.line_no}, the line was:

            {line_info.line}
        """
        exit_error(msg)


def check_dimension_number_is_integer_or_exit(dimension_number, line_info):
    if not is_int(dimension_number):
        msg = f"""
            a #INAME line in an xeasy file must have an integer as its seconds field i got
            {dimension_number}
            the expected format is #INAME <DIMENSION-NUMBER> <AXIS-CODE> [e.g #INAME 1 HN]
            i got the following in ther file {line_info.file_name} at line {line_info.line_no}

            {line_info.line}
        """
        exit_error(msg)


def _check_iname_fields_length_or_exit(fields, line_info):
    if len(fields) != 3:
        msg = f"""
                        a #INAME line in an xeasy file must have 3 fields i got {len(fields)}
                        the expected format is #INAME <DIMENSION-NUMBER> <AXIS-CODE> [e.g #INAME 1 HN]
                        i got the following in ther file {line_info.file_name} at line {line_info.line_no}

                        {line_info.line}
                    """
        exit_error(msg)


def check_number_dimensions_or_exit(line_info, number_dimensions):
    if not is_int(number_dimensions):
        msg = f"""
                        The number of dimensions found in the file {line_info.file_name} doesn't appear to be an integer
                        I got {number_dimensions} at line {line_info.line_no} the value for the line was
                        {line_info.line}
                    """
        exit_error(msg)


def _check_for_peaks_magic_or_exit(line_info):
    if line_info.line != XEASY_PEAK_MAGIC:
        msg = f"""
                    an xeasy peak files should start with {XEASY_PEAK_MAGIC} but i got:

                    {line_info.line}

                    at line {line_info.line_no} in {line_info.file_name}
                """
        exit_error(msg)


def exit_if_wrong_field_count_sequence(fields: List[str], line_info: LineInfo):
    if len(fields) != 2:
        msg = f"""
                in file {line_info.file_name} at line {line_info.line_no} there should be 2 columns i got {len(fields)}
                the line should be of the form 'HIS A1' [<RESIDUE-NAME> <CHAIN-RESIDUE-NUMBER>] but was

                {line_info.line}
            """

        exit_error(msg)


def exit_if_wrong_field_count_peaks(fields: List[str], line_info: LineInfo):
    if len(fields) != 2:
        msg = f"""
                in file {line_info.file_name} at line {line_info.line_no} there should be 2 columns i got {len(fields)}
                the line should be of the form 'HIS A1' [<RESIDUE-NAME> <CHAIN-RESIDUE-NUMBER>] but was

                {line_info.line}
            """

        exit_error(msg)



from pathlib import Path
from typing import List

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    NEF_MOLECULAR_SYSTEM,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import sequence_from_entry, sequence_to_nef_frame
from nef_pipelines.lib.util import (
    exit_error,
    get_text_from_file_or_exit,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.xplor import import_app
from nef_pipelines.transcoders.xplor.psf_lib import parse_xplor_PSF

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# todo add comment to other sequences etc
@import_app.command()
def sequence(
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    entry_name: str = typer.Option("xplor", help="a name for the entry if required"),
    input_path: Path = typer.Option(
        None,
        metavar="|PIPE|",
        help="file to read NEF data from default is stdin '-'",
    ),
    quiet: bool = typer.Option(
        False, "--quiet", "-q", help="if set don't output the nef stream"
    ),
    file_paths: List[Path] = typer.Argument(
        None, help="the file to read", metavar="<XPLOR-PSF-FILES>"
    ),
):
    """- convert xplor psf to nef sequence"""

    # NOTEs
    # 1. we read in the residues and add them to the current sequence frame if there is one
    # 2. this means if you have a sequence frame already you may want to delete it if it clashes...

    if len(file_paths) == 0:
        exit_error("no psf files provided to read sequences from")

    if input_path:
        entry = read_entry_from_file_or_stdin_or_exit_error(input_path)
    else:
        entry = Entry.from_scratch(entry_id=entry_name)

    sequence_residues = OrderedSet(sequence_from_entry(entry))

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    no_chain_ends = parse_comma_separated_options(no_chain_ends)

    xplor_psf_residues = []
    for file_path in file_paths:
        text = get_text_from_file_or_exit(file_path)
        xplor_psf_residues = parse_xplor_PSF(text)

    if len(xplor_psf_residues) == 0:
        exit_error(f"no residues read from {input_path}")

    sequence_residues.update(xplor_psf_residues)

    sequence_frame = sequence_to_nef_frame(
        sequence_residues, no_chain_starts, no_chain_ends
    )

    if NEF_MOLECULAR_SYSTEM in entry:
        entry.remove_saveframe(NEF_MOLECULAR_SYSTEM)
    entry.add_saveframe(sequence_frame)

    if not quiet:
        print(entry)



import string
import sys
from collections import Counter
from enum import IntEnum, auto
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Tuple, Union

import typer
from ordered_set import OrderedSet
from parse import compile as parse_compile
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.nef_frames_lib import NEF_PIPELINES_NAMESPACE
from nef_pipelines.lib.nef_lib import (
    NEF_RELAXATION_VERSION,
    UNUSED,
    SelectionType,
    add_frames_to_entry,
    create_nef_save_frame,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.util import (
    exit_error,
    is_float,
    is_int,
    parse_comma_separated_options,
    strings_to_tabulated_terminal_sensitive,
    warn,
)
from nef_pipelines.tools.series import series_app

FRAMES_AND_TIMES_HELP = "the list of frame selectors and their values"


class RelaxationExperimentType(StrEnum):
    AUTO_RELAXATION = auto()
    DIPOLE_CSA_CROSS_CORRELATIONS = auto()
    DIPOLE_DIPOLE_CROSS_CORRELATIONS = auto()
    DIPOLE_DIPOLE_RELAXATION = auto()
    HETERONUCLEAR_NOES = auto()
    HETERONUCLEAR_R1_RELAXATION = auto()
    HETERONUCLEAR_R1RHO_RELAXATION = auto()
    HETERONUCLEAR_R2_RELAXATION = auto()
    H_EXCHANGE_PROTECTION_FACTORS = auto()
    H_EXCHANGE_RATES = auto()
    HOMONUCLEAR_NOES = auto()
    HETRONUCLEAR_T1_NOE = auto()
    CPMG = auto()
    CEST = auto()
    OTHER = auto()
    T1 = HETERONUCLEAR_R1_RELAXATION
    T2 = HETERONUCLEAR_R2_RELAXATION
    T1RHO = HETERONUCLEAR_R1RHO_RELAXATION
    R1 = HETERONUCLEAR_R1_RELAXATION
    R2 = HETERONUCLEAR_R2_RELAXATION
    NOE = HETERONUCLEAR_NOES


EXPERIMENT_TYPE_TO_SERIES_VARIABLE_TYPE = {
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "time",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "time",
    RelaxationExperimentType.HETRONUCLEAR_T1_NOE: "time",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "saturation",
    RelaxationExperimentType.CPMG: "cycles",
    RelaxationExperimentType.CEST: "offset",
    RelaxationExperimentType.OTHER: "unknown",
}

EXPERIMENT_TYPE_TO_DATA_VARIABLE_TYPE = {
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "intensity",
    RelaxationExperimentType.HETRONUCLEAR_T1_NOE: "intensity",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "intensity",
    RelaxationExperimentType.CPMG: "intensity",
    RelaxationExperimentType.CEST: "intensity",
    RelaxationExperimentType.OTHER: "unknown",
}

DATA_VARIABLE_TYPE_TO_DATA_VARIABLE_UNIT = {
    "intensity": UNUSED,
    "volume": UNUSED,
    "unknown": UNUSED,
}

SERIES_VARIABLE_TYPE_TO_SERIES_VARIABLE_UNIT = {
    "time": "s",
    "saturation": "fraction",
    "cycles": UNUSED,
    "offset": "hz",
}

EXPERIMENT_TO_IDENTIFIER = {
    RelaxationExperimentType.AUTO_RELAXATION: "auto_relaxation",
    RelaxationExperimentType.DIPOLE_CSA_CROSS_CORRELATIONS: "dipole_CSA_cross_correlations",
    RelaxationExperimentType.DIPOLE_DIPOLE_CROSS_CORRELATIONS: "dipole_dipole_cross_correlations",
    RelaxationExperimentType.DIPOLE_DIPOLE_RELAXATION: "dipole_dipole_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_NOES: "heteronuclear_NOEs",
    RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION: "heteronuclear_R1_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION: "heteronuclear_R1rho_relaxation",
    RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION: "heteronuclear_R2_relaxation",
    RelaxationExperimentType.H_EXCHANGE_PROTECTION_FACTORS: "H_exchange_protection_factors",
    RelaxationExperimentType.H_EXCHANGE_RATES: "H_exchange_rates",
    RelaxationExperimentType.HOMONUCLEAR_NOES: "homonuclear_NOEs",
    RelaxationExperimentType.CPMG: "CPMG",
    RelaxationExperimentType.CEST: "CEST",
    RelaxationExperimentType.OTHER: "other",
}

NAME_TO_EXPERIMENT_TYPE = {
    "T1": RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION,
    "T2": RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION,
    "T1rho": RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION,
    "T1p": RelaxationExperimentType.HETERONUCLEAR_R1RHO_RELAXATION,
    "R1": RelaxationExperimentType.HETERONUCLEAR_R1_RELAXATION,
    "R2": RelaxationExperimentType.HETERONUCLEAR_R2_RELAXATION,
    "NOE": RelaxationExperimentType.HETERONUCLEAR_NOES,
    "CPMG": RelaxationExperimentType.CPMG,
    "REX": RelaxationExperimentType.CPMG,
    "CEST": RelaxationExperimentType.CEST,
    "DEST": RelaxationExperimentType.CEST,
}


class SaturationState(IntEnum):
    SATURATED = 1
    UNSATURATED = 0


STR_TO_SATURATION_STATE = {
    "saturated": SaturationState.SATURATED,
    "unsaturated": SaturationState.UNSATURATED,
    "on": SaturationState.SATURATED,
    "off": SaturationState.UNSATURATED,
    "true": SaturationState.SATURATED,
    "false": SaturationState.UNSATURATED,
    "yes": SaturationState.SATURATED,
    "no": SaturationState.UNSATURATED,
    "1": SaturationState.SATURATED,
    "0": SaturationState.UNSATURATED,
}


@series_app.command()
def build(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    timings: List[str] = typer.Option(
        None,
        help="the list of relaxation times or values",
    ),
    input_unit: str = typer.Option(
        None,
        "--unit",
        help="the unit of the relaxation times",
    ),
    name: str = typer.Option(
        None,
        help="the name of the series, if not provided it will be guessed from the frames",
    ),
    experiment_type: str = typer.Option(None, help="the type of the experiment"),
    display_parsed_values: bool = typer.Option(False, help="display the parsed values"),
    frame_selectors: List[str] = typer.Argument(
        None,
        help=FRAMES_AND_TIMES_HELP,
    ),
):
    """- build a NEF data series from a set of spectra [alpha]"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    frame_selectors = parse_comma_separated_options(frame_selectors)

    _exit_if_no_frame_selectors(frame_selectors)

    frames_by_selector = _select_relaxation_frames_by_selector_or_exit_if_other(
        entry, frame_selectors
    )

    frame_timings_and_units_by_selector = {}
    for selector, frames in frames_by_selector.items():
        frame_timings_and_units_by_selector[selector] = (
            _get_timings_and_units_for_frames(
                frames,
                [
                    selector,
                ],
                timings,
                input_unit,
                display_parsed_values,
            )
        )

    for (
        selector,
        frame_timings_and_units,
    ) in frame_timings_and_units_by_selector.items():
        frames_and_timings, unit = (
            _ensure_units_consistent_and_get_unit_and_timings_or_exit(
                frame_timings_and_units
            )
        )

    frames_and_timings = {}
    for frame_timings_and_units in frame_timings_and_units_by_selector.values():
        frames_and_timings.update(frame_timings_and_units)

    # TODO aren't these the same? add a name and expt attribute to the template and only use this notherwise
    if not name:
        name = _make_series_name(entry, frames_and_timings)

    if not experiment_type:
        experiment_type = _guess_experiment_type(name)

    entry = pipe(entry, frames_and_timings, unit, name, experiment_type)

    print(entry)


def pipe(
    entry: Entry,
    frame_timings: Dict[Tuple[str, int], Union[int, float, bool]],
    unit: str,
    name: str,
    experiment_type: str,
) -> Entry:

    series_frame = create_nef_save_frame(f"{NEF_PIPELINES_NAMESPACE}_series_list", name)

    if experiment_type == "unknown":
        series_variable_type = "unknown"
    else:
        series_variable_type = EXPERIMENT_TYPE_TO_SERIES_VARIABLE_TYPE[
            RelaxationExperimentType[experiment_type]
        ]

    if series_variable_type == "unknown":
        series_variable_unit = "unknown"
    else:

        series_variable_unit = SERIES_VARIABLE_TYPE_TO_SERIES_VARIABLE_UNIT[
            series_variable_type
        ]

    series_variable_unit = (
        unit if unit != series_variable_unit else series_variable_unit
    )

    if experiment_type == "unknown":
        data_value_type = "unknown"
    else:
        data_value_type = EXPERIMENT_TYPE_TO_DATA_VARIABLE_TYPE[
            RelaxationExperimentType[experiment_type]
        ]
    data_value_unit = DATA_VARIABLE_TYPE_TO_DATA_VARIABLE_UNIT[data_value_type]

    if experiment_type == "unknown":
        experiment_type = RelaxationExperimentType.OTHER
    else:
        experiment_type = EXPERIMENT_TO_IDENTIFIER[
            RelaxationExperimentType[experiment_type]
        ]
    tags = [
        ("experiment_type", experiment_type),
        ("series_variable_type", series_variable_type),
        ("series_variable_unit", series_variable_unit),
        ("data_variable_type", series_variable_type),
        ("data_variable_unit", series_variable_unit),
        ("data_value_type", data_value_type),
        ("data_value_unit", data_value_unit),
        (
            "version",
            NEF_RELAXATION_VERSION,
        ),
        ("comment", UNUSED),
    ]
    series_frame.add_tags(tags)

    # TODO: pynmstar doesn't catch prepended spaces here it just gives an obscure error report!
    series_experiment_loop = Loop.from_scratch(
        f"{NEF_PIPELINES_NAMESPACE}_series_experiment"
    )
    series_frame.add_loop(series_experiment_loop)
    series_experiment_loop.add_tag(
        [
            "nmr_spectrum_id",
            "reference_experiment",
            "combination_id",
            "pseudo_dimension",
            "pseudo_dimension_point",
            "series_variable",
            "series_variable_error",
        ]
    )

    loop_data = []
    for frame_key, (value, unit) in frame_timings.items():
        frame_name, _ = frame_key

        value = _scale_value_by_unit(value, unit)

        loop_data.append(
            {
                "nmr_spectrum_id": frame_name,
                "reference_experiment": "false",
                "combination_id": UNUSED,
                "pseudo_dimension": UNUSED,
                "pseudo_dimension_point": UNUSED,
                "series_variable": value,
                "series_variable_error": UNUSED,
            }
        )

    series_experiment_loop.add_data(loop_data)

    add_frames_to_entry(
        entry,
        [
            series_frame,
        ],
    )

    return entry


def _guess_experiment_type(name):
    upper_name = name.upper()
    for experiment_name in NAME_TO_EXPERIMENT_TYPE:
        if experiment_name in upper_name and upper_name in NAME_TO_EXPERIMENT_TYPE:
            result = NAME_TO_EXPERIMENT_TYPE[upper_name]
            break
        else:
            result = "unknown"
    return result


def _make_series_name(entry, frames_and_timings):
    result = None
    for relaxation_type in NAME_TO_EXPERIMENT_TYPE:
        do_break = False
        for name, _ in frames_and_timings.keys():
            name = name.upper()
            if relaxation_type in name.upper():
                start_index = name.index(relaxation_type)
                end_index = start_index + len(relaxation_type)
                found_relaxation_type = name[start_index:end_index]

                result = found_relaxation_type
                do_break = True
                break
        if do_break:
            break

    if not result:
        result = "unknown_type"

    bad_frame_name = False
    for frame in entry:
        if frame.name == result:
            bad_frame_name = True
            break

    if bad_frame_name:
        base_result = result
        for i in range(1, 100):
            result = f"series`{i}`"
            bad_frame_name = False
            for frame in entry:
                if frame.name == result:
                    bad_frame_name = True
                    break

    if bad_frame_name:
        _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result)

    return result


def _exit_if_too_many_clashing_frame_names(bad_frame_name, base_result):
    if bad_frame_name:
        msg = f"""
            after searching the series of names {base_result}`1` to {base_result}`100` i could not find a name
            that was not already in the entry, choose a name on the command line, you have way too much data!
            more serious report this as a bug...
        """
        exit_error(msg)


def _get_timings_and_units_for_frames(
    frames, frame_selectors, timings, input_unit, display_parsed_values
):

    if timings:
        timings_and_units = _parse_timings(timings)
        units = [timing_and_unit[1] for timing_and_unit in timings_and_units]

        _exit_if_input_unit_defined_and_any_unit_parsed(
            input_unit, units, timings_and_units
        )

        if input_unit:
            timings_and_units = [
                (timing, input_unit) for timing, _ in timings_and_units
            ]

        _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
            frames, timings_and_units
        )

        frame_timings_and_units = {
            (frame.name, id(frame)): timing_and_unit
            for frame, timing_and_unit in zip(frames, timings_and_units)
        }
    else:
        _exit_if_frame_selector_has_no_timing_selector(frame_selectors)
        frame_timings_and_units = _parse_timings_from_frames_or_exit(
            frames, frame_selectors[0], display_parsed_values
        )

    return frame_timings_and_units


def _exit_if_input_unit_defined_and_any_unit_parsed(
    input_unit, units, timings_and_units
):
    if input_unit and any(unit != "" for unit in units):
        units_and_timings_strings = [
            f"{unit}{timing}" for unit, timing in timings_and_units if unit
        ]
        msg = f"""
                you cannot provide a unit on the command line [--unit] and units in the timings [--timings]
                at the same time command line unit was {input_unit} and the units and timings were
                {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}

                """
        exit_error(msg)


def _exit_error_it_number_of_timings_doesnt_match_number_of_frames(
    frames, timings_and_units
):
    if len(timings_and_units) != len(frames):
        msg = f"the number of timings {len(timings_and_units)} does not match the number of frames {len(frames)}"
        exit_error(msg)


def _exit_if_frame_selector_has_no_timing_selector(frame_selectors):

    reason = None

    if (len(frame_selectors)) < 1:
        reason = "no frames were selected"
    else:
        for i, frame_selector in enumerate(frame_selectors, start=1):
            if "{var}" not in frame_selector:
                reason = f"frame selector {i}: {frame_selector} doesn't contain a {{var}} place holder"

    if reason:
        msg = f"""\
            timings not parsed because:

            {reason}

            to select timings you must have one or more frame selector each with
            a {{var}} placeholder. I got {len(frame_selectors)} frame selectors and they were:

            """
        msg = dedent(msg)
        frame_elector_strings = "\n".join(frame_selectors)
        msg += frame_elector_strings
        exit_error(msg)


def _select_relaxation_frames_by_selector_or_exit_if_other(
    entry: Entry, frame_selectors: List[str]
) -> List[Saveframe]:
    """Select relaxation frames that are also spectrum frames using multiple selectors.

    Args:
        entry: NEF entry to search in
        frame_selectors: List of frame frame_selector patterns

    Returns:
        List of spectrum frames that match the selectors (no duplicates)
    """

    # Get all spectrum frames
    spectrum_frames = select_frames(entry, "nef_nmr_spectrum", SelectionType.CATEGORY)
    spectrum_frame_ids = OrderedSet(
        [
            (spectrum_frame.name, id(spectrum_frame))
            for spectrum_frame in spectrum_frames
        ]
    )

    # Process selectors with variable substitution and rememebr originals
    processed_selectors_and_original_selectors = {}
    for frame_selector in frame_selectors:

        processed_selector = frame_selector.replace("{var}", "*").replace("{}", "*")
        processed_selectors_and_original_selectors[processed_selector] = frame_selector

    # Collect all frames matching any frame_selector (use OrderedSet to avoid duplicates)
    frame_ids_by_selector = {}
    for (
        processed_frame_selector,
        frame_selector,
    ) in processed_selectors_and_original_selectors.items():
        selected_frame_ids = OrderedSet()
        frame_ids_by_selector[frame_selector] = selected_frame_ids

        matching_frames = select_frames(
            entry, processed_frame_selector, SelectionType.NAME
        )
        for frame in matching_frames:
            selected_frame_ids.add((frame.name, id(frame)))

    # Filter to only include spectrum frames (intersection)
    for frame_selector, selected_frame_ids in frame_ids_by_selector.items():
        frame_ids_by_selector[frame_selector] = spectrum_frame_ids.intersection(
            selected_frame_ids
        )

    # Check for non-spectrum frames and error if found (preserve original behavior)
    for frame_selector, valid_frame_ids in frame_ids_by_selector.items():
        non_spectrum_frame_ids = frame_ids_by_selector[frame_selector].difference(
            valid_frame_ids
        )
        if non_spectrum_frame_ids:
            non_spectrum_frame_names = "\n".join(
                [frame_name for frame_name, _ in non_spectrum_frame_ids]
            )
            _exit_if_some_frames_are_not_spectra(
                frame_selector, non_spectrum_frame_names
            )

    # Convert back to frame objects
    valid_frames_by_selector = {}
    for frame_selector, valid_frame_ids in frame_ids_by_selector.items():
        valid_frames_by_selector[frame_selector] = [
            entry.get_saveframe_by_name(frame_name) for frame_name, _ in valid_frame_ids
        ]

    return valid_frames_by_selector


def _exit_if_some_frames_are_not_spectra(frame_selector, non_spectrum_frame_names):
    msg = f"""
            using the frame selector {frame_selector}
            some of the selected frames are not spectrum frames:
            {non_spectrum_frame_names}
            """
    exit_error(msg)


def _parse_timings_from_frames_or_exit(
    save_frames: Dict[Tuple[str, int], List[Saveframe]],
    frame_selector: str,
    display_parsed_values: bool,
) -> Dict[Tuple[str, int], List[Tuple[Union[int, float, bool], str]]]:

    if "{var}" not in frame_selector:
        msg = f"no timing selector placeholder {{var}} provided in frame selector {frame_selector}"
        exit_error(msg)

    frame_selector = frame_selector.replace("{var}", "{var}")
    frame_selector = frame_selector.replace("*", "{:opt}")

    parser = parse_compile(frame_selector)

    parsed_values = []

    frame_value_and_units = {}
    for frame in save_frames:
        frame_key = frame.name, id(frame)
        result = parser.search(frame.name)

        value, unit = None, None
        if result and "var" in result:
            raw_value = result["var"]
            raw_value = raw_value.replace("_", ".")
            raw_value = raw_value.replace("-", ".")

            value, unit = _parse_value_and_unit(raw_value)

        if display_parsed_values:
            parsed_values.append([frame.name, raw_value, value, unit])

        if value is None:
            _exit_if_no_variable_found(frame, frame_selector)

        if not (is_float(value) or value.lower() in "on off true false yes no"):
            msg = f"the variable {value} in frame {frame.name} is not a number or a boolean"
            exit_error(msg)

        frame_value_and_units[frame_key] = value, unit

    if display_parsed_values:
        print(
            tabulate(parsed_values, headers=["spectrum", "raw value", "value", "unit"]),
            file=sys.stderr,
        )
        sys.exit(0)

    return frame_value_and_units


def _exit_if_no_variable_found(frame, frame_selector):
    msg = f"no variable found in frame {frame.name} with the selector {frame_selector}"
    exit_error(msg)


def _parse_timings(timings: List[str]) -> List[Tuple[Union[int, float, bool], str]]:
    timings = parse_comma_separated_options(timings)
    parsed_timings = []
    for timing in timings:
        value, unit = _parse_value_and_unit(timing)
        parsed_timings.append((value, unit))
    return parsed_timings


def _parse_value_and_unit(value: str) -> Tuple[Union[bool, int, float, None]]:

    unit = value.lstrip(string.digits + ".eE")
    len_digits = len(value) - len(unit)
    value = value[:len_digits]

    lower_unit = unit.lower()
    if value and unit:
        pass
    elif not value and lower_unit in STR_TO_SATURATION_STATE:
        unit = "bool"
        value = STR_TO_SATURATION_STATE[lower_unit]
    elif not unit:
        if is_int(value):
            value = int(value)
        elif is_float(value):
            value = float(value)
        unit = ""
    else:
        unit = None
        value = None

    return value, unit


def _exit_if_no_frame_selectors(frame_selectors):
    if len(frame_selectors) == 0:
        msg = "you must select some frames!"
        exit_error(msg)


def _ensure_units_consistent_and_get_unit_and_timings_or_exit(frames_timings_and_units):

    _exit_units_not_consistent(frames_timings_and_units)

    _exit_if_numbers_and_bools_are_mixed(frames_timings_and_units)

    float_counter = Counter()
    int_counter = Counter()
    for frame, (value, unit) in frames_timings_and_units.items():
        if isinstance(value, int):
            int_counter[value] += 1
        elif isinstance(value, float):
            float_counter[value] += 1

    if int_counter and float_counter:
        for frame_key, (value, unit) in frames_timings_and_units.items():
            value = float(value)

            frames_timings_and_units[frame_key] = value, unit

    unit = list(frames_timings_and_units.values())[0][1]
    frames_and_timings = {
        frame_key: value for frame_key, (value, _) in frames_timings_and_units.items()
    }

    return frames_and_timings, unit


def _exit_if_numbers_and_bools_are_mixed(frame_timings_and_units):

    counter_numbers = Counter()
    couter_booleans = Counter()

    for frame, (value, unit) in frame_timings_and_units.items():
        if isinstance(value, bool):
            couter_booleans[value] += 1
        elif is_float(value):
            counter_numbers[value] += 1

    if counter_numbers and couter_booleans:
        numbers_strings = ", ".join(counter_numbers.keys())
        booleans_strings = ", ".join(couter_booleans.keys())
        msg = f"""
                you cannot have both numbers and booleans in the values
                the booleans were
                {booleans_strings}
                the numbers were
                {numbers_strings}
            """
        exit_error(msg)


def _exit_units_not_consistent(frame_timings_and_units):

    unit_counter = Counter()
    for frame, (value, unit) in frame_timings_and_units.items():
        unit_counter[unit] += 1

    if len(unit_counter) > 1:
        unit_strings = ", ".join(unit_counter.keys())
        msg = f"""
            the units of the timings are not consistent, the units found were
            {unit_strings}
            """
        exit_error(msg)


def _exit_if_units_are_not_compatible(frame_timings_and_units):
    units = {unit for timing, unit in frame_timings_and_units.values()}
    if len(units) > 1:
        units_and_timings_strings = [
            f"{unit}{timing}" for timing, unit in frame_timings_and_units.values()
        ]
        msg = f"""
                the units of the timings are not compatible they are {units}
                the timings and units were
                {strings_to_tabulated_terminal_sensitive(units_and_timings_strings)}
                """
        exit_error(msg)


def _scale_value_by_unit(value, unit):
    if is_float(value):
        value = float(value)
    if not unit or unit == "bool":
        pass
    elif unit.lower() in ("us", "ms"):
        if unit.lower() == "ms":
            value *= 1e-3
        elif unit.lower() == "us":
            value *= 1e-6
    else:
        msg = f""" scaling by unit {unit} not implemented [value: {value}], ignoring!"""
        warn(msg)

    return round(value, 6)



import os
from textwrap import dedent

import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrview.exporters.sequences import sequences

app = typer.Typer()
app.command()(sequences)

INPUT_MULTI_CHAIN_NEF = read_test_data("multi_chain.nef", __file__)


# noinspection PyUnusedLocal
def test_multi_chain():

    result = run_and_report(app, ["-o", "-"], input=INPUT_MULTI_CHAIN_NEF)

    EXPECTED = """\
        ------------- A.seq -------------
        his 3
        met
        ------------- B.seq -------------
        arg 5
        gln
        ------------- C.seq -------------
        pro 7
        ---------------------------------
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain_template():

    result = run_and_report(
        app, ["-o", "-test_{chain_code}.seq"], input=INPUT_MULTI_CHAIN_NEF
    )

    EXPECTED = """\
        ------------- test_A.seq -------------
        his 3
        met
        ------------- test_B.seq -------------
        arg 5
        gln
        ------------- test_C.seq -------------
        pro 7
        --------------------------------------
    """

    assert_lines_match(EXPECTED, result.stdout)


def test_multi_chain_bad_selector():

    result = run_and_report(
        app, ["-o", "-", "AAAA"], input=INPUT_MULTI_CHAIN_NEF, expected_exit_code=1
    )

    assert "the chain code" in result.stdout
    assert "not in the molecular systems chain codes" in result.stdout
    assert "AAAA" in result.stdout


def test_bad_template():

    result = run_and_report(
        app, ["-o", "test.seq", "A"], input=INPUT_MULTI_CHAIN_NEF, expected_exit_code=1
    )

    assert "the file name template" in result.stdout
    assert "does not contain the string" in result.stdout
    assert "{chain_code}" in result.stdout


def test_multiple_output_files_disk(
    tmp_path,
):

    os.chdir(tmp_path)

    run_and_report(app, [], input=INPUT_MULTI_CHAIN_NEF)

    EXPECTED_FILES = "A.seq", "B.seq", "C.seq"

    for file_name in EXPECTED_FILES:
        assert (tmp_path / file_name).is_file()

    EXPECTED = {
        "A.seq": """\
            his 3
            met
        """,
        "B.seq": """\
            arg 5
            gln
        """,
        "C.seq": """\
            pro 7
        """,
    }

    for file_name in EXPECTED_FILES:
        with open(file_name) as fp:
            text = fp.read()

            assert text == dedent(EXPECTED[file_name])



import sys
from pathlib import Path

import typer

from nef_pipelines.lib.nef_lib import read_entry_from_stdin_or_exit
from nef_pipelines.tools.entry import entry_app


# noinspection PyUnusedLocal
@entry_app.command()
def rename(
    input: Path = typer.Option(
        None,
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="read NEF data from a file instead of stdin",
    ),
    name: str = typer.Argument(
        None,
        help="the new name for the entry",
    ),
):
    """- rename the current entry"""

    entry = read_entry_from_stdin_or_exit()

    if name is not None:
        entry.entry_id = name

    if not sys.stdout.isatty():
        print(entry)
